__author__ = "Simon Nilsson"

import io
import os
from typing import Dict, List

import cv2
import matplotlib.pyplot as plt
import numpy as np
import PIL

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.enums import Formats
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.lookups import get_color_dict
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class DistancePlotterSingleCore(ConfigReader, PlottingMixin):
    """
    Class for visualizing the distance between two pose-estimated body-parts (e.g., two animals) through line
    charts. Results are saved as individual line charts, and/or videos of line charts.

    .. note::
       For better runtime, use :meth:`simba.plotting.distance_plotter_mp.DistancePlotterMultiCore`.
       `GitHub tutorial/documentation <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. image:: _static/img/distance_plot.png
       :width: 300
       :align: center

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter bool frame_setting: If True, creates individual frames.
    :parameter bool video_setting: If True, creates videos

    :examples:

    >>> style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8, 'opacity': 0.5}
    >>> line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
    >>> distance_plotter = DistancePlotterSingleCore(config_path=r'MyProjectConfig', files_found=['test/two_c57s/project_folder/csv/outlier_corrected_movement_location/Video_1.csv'], frame_setting=False, video_setting=True, final_img=True)
    >>> distance_plotter.run()
    """

    def __init__(
        self,
        config_path: str,
        frame_setting: bool,
        video_setting: bool,
        final_img: bool,
        files_found: List[str],
        style_attr: Dict[str, int],
        line_attr: Dict[int, list],
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        (
            self.video_setting,
            self.frame_setting,
            self.files_found,
            self.style_attr,
            self.line_attr,
            self.final_img,
        ) = (
            video_setting,
            frame_setting,
            files_found,
            style_attr,
            line_attr,
            final_img,
        )
        if (not frame_setting) and (not video_setting) and (not self.final_img):
            raise NoSpecifiedOutputError(
                "Please choice to create frames and/or video distance plots"
            )
        self.colors_dict = get_color_dict()
        if not os.path.exists(self.line_plot_dir):
            os.makedirs(self.line_plot_dir)
        check_if_filepath_list_is_empty(
            filepaths=self.outlier_corrected_paths,
            error_msg="SIMBA ERROR: Zero files found in the project_folder/csv/machine_results directory. Create classification results before visualizing distances.",
        )
        print(f"Processing {str(len(self.files_found))} videos...")

    def run(self):
        """
        Creates line charts. Results are stored in the `project_folder/frames/line_plot` directory

        Returns
        ----------
        None
        """

        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            self.data_df = read_df(file_path, self.file_type)
            distance_arr = np.full(
                (len(self.data_df), len(self.line_attr.keys())), np.nan
            )
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            for distance_cnt, data in enumerate(self.line_attr.values()):
                distance_arr[:, distance_cnt] = (
                    np.sqrt(
                        (self.data_df[data[0] + "_x"] - self.data_df[data[1] + "_x"])
                        ** 2
                        + (self.data_df[data[0] + "_y"] - self.data_df[data[1] + "_y"])
                        ** 2
                    )
                    / self.px_per_mm
                ) / 10
            if self.video_setting:
                self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
                self.video_save_path = os.path.join(
                    self.line_plot_dir, self.video_name + ".mp4"
                )
                writer = cv2.VideoWriter(
                    self.video_save_path,
                    self.fourcc,
                    self.fps,
                    (self.style_attr["width"], self.style_attr["height"]),
                )
            if self.frame_setting:
                self.save_video_folder = os.path.join(
                    self.line_plot_dir, self.video_name
                )
                if not os.path.exists(self.save_video_folder):
                    os.makedirs(self.save_video_folder)

            distance_arr = np.nan_to_num(distance_arr, nan=0.0)
            if self.final_img:
                self.final_img_path = os.path.join(
                    self.line_plot_dir, self.video_name + "_final_img.png"
                )
                self.make_distance_plot(
                    data=distance_arr,
                    line_attr=self.line_attr,
                    style_attr=self.style_attr,
                    fps=self.fps,
                    save_path=self.final_img_path,
                    save_img=True,
                )

            if self.video_setting or self.frame_setting:
                if self.style_attr["y_max"] == "auto":
                    max_y = np.amax(distance_arr)
                else:
                    max_y = float(self.style_attr["y_max"])
                y_ticks_locs = y_lbls = np.round(np.linspace(0, max_y, 10), 2)
                for i in range(distance_arr.shape[0]):
                    for j in range(distance_arr.shape[1]):
                        color = self.colors_dict[self.line_attr[j][-1]][::-1]
                        color = tuple(x / 255 for x in color)
                        plt.plot(
                            distance_arr[0:i, j],
                            color=color,
                            linewidth=self.style_attr["line width"],
                            alpha=self.style_attr["opacity"],
                        )

                    x_ticks_locs = x_lbls = np.round(np.linspace(0, i, 5))
                    x_lbls = np.round((x_lbls / self.fps), 1)
                    plt.ylim(0, max_y)

                    plt.xlabel("time (s)")
                    plt.ylabel("distance (cm)")
                    plt.xticks(
                        x_ticks_locs,
                        x_lbls,
                        rotation="horizontal",
                        fontsize=self.style_attr["font size"],
                    )
                    plt.yticks(
                        y_ticks_locs, y_lbls, fontsize=self.style_attr["font size"]
                    )
                    plt.suptitle(
                        "Animal distances",
                        x=0.5,
                        y=0.92,
                        fontsize=self.style_attr["font size"] + 4,
                    )

                    self.buffer_ = io.BytesIO()
                    plt.savefig(self.buffer_, format="png")
                    self.buffer_.seek(0)
                    img = PIL.Image.open(self.buffer_)
                    img = np.uint8(cv2.cvtColor(np.asarray(img), cv2.COLOR_RGB2BGR))
                    self.buffer_.close()
                    plt.close()

                    img = cv2.resize(
                        img, (self.style_attr["width"], self.style_attr["height"])
                    )

                    if self.frame_setting:
                        frame_save_path = os.path.join(
                            self.save_video_folder, str(i) + ".png"
                        )
                        cv2.imwrite(frame_save_path, img)
                    if self.video_setting:
                        writer.write(img)
                    print(
                        "Distance frame: {} / {}. Video: {} ({}/{})".format(
                            str(i + 1),
                            str(len(self.data_df)),
                            self.video_name,
                            str(file_cnt + 1),
                            len(self.files_found),
                        )
                    )

                if self.video_setting:
                    writer.release()
                video_timer.stop_timer()
                print(
                    "Distance plot for video {} saved (elapsed time: {}s)...".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )
        self.timer.stop_timer()
        stdout_success(
            msg="All distance visualizations created in project_folder/frames/output/line_plot directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


#
# style_attr = {'width': 640,
#               'height': 480,
#               'line width': 6,
#               'font size': 8,
#               'y_max': 'auto',
#               'opacity': 0.9}
# line_attr = {0: ['Center_1', 'Center_2', 'Green'], 1: ['Ear_left_2', 'Ear_left_1', 'Red']}
#
# test = DistancePlotterSingleCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                  frame_setting=False,
#                                  video_setting=True,
#                                  style_attr=style_attr,
#                                  final_img=True,
#                                  files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                  line_attr=line_attr)
# test.create_distance_plot()

# style_attr = {'width': 640, 'height': 480, 'line width': 6, 'font size': 8}
# line_attr = {0: ['Termite_1_Head_1', 'Termite_1_Thorax_1', 'Dark-red']}

# test = DistancePlotterSingleCore(config_path=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini',
#                                  frame_setting=False,
#                        video_setting=True,
#                        style_attr=style_attr,
#                        files_found=['/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/csv/outlier_corrected_movement_location/termites_1.csv'],
#                        line_attr=line_attr)
# test.create_distance_plot()



__author__ = "Simon Nilsson"

import multiprocessing
import os
from tkinter import *

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.heat_mapper_location import HeatmapperLocationSingleCore
from simba.plotting.heat_mapper_location_mp import \
    HeatMapperLocationMultiprocess
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, DropDownMenu
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.enums import Formats, Keys, Links, Paths
from simba.utils.read_write import get_file_name_info_in_directory


class HeatmapLocationPopup(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="HEATMAPS: LOCATION")
        ConfigReader.__init__(self, config_path=config_path)
        self.data_path = os.path.join(self.project_path, Paths.OUTLIER_CORRECTED.value)
        self.files_found_dict = get_file_name_info_in_directory(
            directory=self.data_path, file_type=self.file_type
        )
        check_if_filepath_list_is_empty(
            filepaths=list(self.files_found_dict.keys()),
            error_msg="SIMBA ERROR: Zero files found in the project_folder/csv/outlier_corrected_movement_location directory. ",
        )
        max_scales = list(np.linspace(5, 600, 5))
        max_scales.insert(0, "Auto-compute")
        self.style_settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="STYLE SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.HEATMAP_LOCATION.value,
        )
        self.palette_dropdown = DropDownMenu(
            self.style_settings_frm, "Palette:", self.palette_options, "16"
        )
        self.shading_dropdown = DropDownMenu(
            self.style_settings_frm, "Shading:", self.shading_options, "16"
        )
        self.bp_dropdown = DropDownMenu(
            self.style_settings_frm, "Body-part:", self.body_parts_lst, "16"
        )
        self.max_time_scale_dropdown = DropDownMenu(
            self.style_settings_frm, "Max time scale (s):", max_scales, "16"
        )
        self.bin_size_dropdown = DropDownMenu(
            self.style_settings_frm,
            "Bin size (mm):",
            self.heatmap_bin_size_options,
            "16",
        )

        self.palette_dropdown.setChoices(self.palette_options[0])
        self.shading_dropdown.setChoices(self.shading_options[0])
        self.bp_dropdown.setChoices(self.body_parts_lst[0])
        self.max_time_scale_dropdown.setChoices(max_scales[0])
        self.bin_size_dropdown.setChoices("80×80")

        self.settings_frm = LabelFrame(
            self.main_frm,
            text="VISUALIZATION SETTINGS",
            font=("Helvetica", 14, "bold"),
            pady=5,
            padx=5,
        )
        self.heatmap_frames_var = BooleanVar()
        self.heatmap_videos_var = BooleanVar()
        self.heatmap_last_frm_var = BooleanVar()
        self.multiprocessing_var = BooleanVar()
        heatmap_frames_cb = Checkbutton(
            self.settings_frm, text="Create frames", variable=self.heatmap_frames_var
        )
        heatmap_videos_cb = Checkbutton(
            self.settings_frm, text="Create videos", variable=self.heatmap_videos_var
        )
        heatmap_last_frm_cb = Checkbutton(
            self.settings_frm,
            text="Create last frame",
            variable=self.heatmap_last_frm_var,
        )
        self.multiprocess_cb = Checkbutton(
            self.settings_frm,
            text="Multiprocess videos (faster)",
            variable=self.multiprocessing_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.multiprocessing_var,
                dropdown_menus=[self.multiprocess_dropdown],
            ),
        )
        self.multiprocess_dropdown = DropDownMenu(
            self.settings_frm, "CPU cores:", list(range(2, self.cpu_cnt)), "12"
        )
        self.multiprocess_dropdown.setChoices(2)
        self.multiprocess_dropdown.disable()

        self.run_frm = LabelFrame(
            self.main_frm,
            text="RUN",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_frm = LabelFrame(
            self.run_frm,
            text="SINGLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_btn = Button(
            self.run_single_video_frm,
            text="Create single video",
            fg="blue",
            command=lambda: self.__create_heatmap_plots(multiple_videos=False),
        )
        self.single_video_dropdown = DropDownMenu(
            self.run_single_video_frm,
            "Video:",
            list(self.files_found_dict.keys()),
            "12",
        )
        self.single_video_dropdown.setChoices(list(self.files_found_dict.keys())[0])
        self.run_multiple_videos = LabelFrame(
            self.run_frm,
            text="MULTIPLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_multiple_video_btn = Button(
            self.run_multiple_videos,
            text="Create multiple videos ({} video(s) found)".format(
                str(len(list(self.files_found_dict.keys())))
            ),
            fg="blue",
            command=lambda: self.__create_heatmap_plots(multiple_videos=False),
        )

        self.style_settings_frm.grid(row=0, sticky=NW)
        self.palette_dropdown.grid(row=0, sticky=NW)
        self.shading_dropdown.grid(row=1, sticky=NW)
        self.bp_dropdown.grid(row=2, sticky=NW)
        self.max_time_scale_dropdown.grid(row=3, sticky=NW)
        self.bin_size_dropdown.grid(row=4, sticky=NW)

        self.settings_frm.grid(row=1, sticky=NW)
        heatmap_frames_cb.grid(row=0, sticky=NW)
        heatmap_videos_cb.grid(row=1, sticky=NW)
        heatmap_last_frm_cb.grid(row=2, sticky=NW)
        self.multiprocess_cb.grid(row=3, column=0, sticky=NW)
        self.multiprocess_dropdown.grid(row=3, column=1, sticky=NW)

        self.run_frm.grid(row=2, sticky=NW)
        self.run_single_video_frm.grid(row=0, sticky=NW)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)

        self.run_multiple_videos.grid(row=1, sticky=NW)
        self.run_multiple_video_btn.grid(row=0, sticky=NW)

        # self.main_frm.mainloop()

    def __create_heatmap_plots(self, multiple_videos: bool):
        if multiple_videos:
            data_paths = list(self.files_found_dict.values())
        else:
            data_paths = [
                self.files_found_dict[self.single_video_dropdown.getChoices()]
            ]

        if self.max_time_scale_dropdown.getChoices() != "Auto-compute":
            max_scale = int(float(self.max_time_scale_dropdown.getChoices()))
        else:
            max_scale = "auto"

        bin_size = int(self.bin_size_dropdown.getChoices().split("×")[0])

        style_attr = {
            "palette": self.palette_dropdown.getChoices(),
            "shading": self.shading_dropdown.getChoices(),
            "max_scale": max_scale,
            "bin_size": bin_size,
        }

        if not self.multiprocessing_var.get():
            heatmapper_clf = HeatmapperLocationSingleCore(
                config_path=self.config_path,
                style_attr=style_attr,
                final_img_setting=self.heatmap_last_frm_var.get(),
                video_setting=self.heatmap_videos_var.get(),
                frame_setting=self.heatmap_frames_var.get(),
                bodypart=self.bp_dropdown.getChoices(),
                files_found=data_paths,
            )

            heatmapper_clf_processor = multiprocessing.Process(heatmapper_clf.run())
            heatmapper_clf_processor.start()

        else:
            heatmapper_clf = HeatMapperLocationMultiprocess(
                config_path=self.config_path,
                style_attr=style_attr,
                final_img_setting=self.heatmap_last_frm_var.get(),
                video_setting=self.heatmap_videos_var.get(),
                frame_setting=self.heatmap_frames_var.get(),
                bodypart=self.bp_dropdown.getChoices(),
                files_found=data_paths,
                core_cnt=int(self.multiprocess_dropdown.getChoices()),
            )

            heatmapper_clf.run()


# _ = HeatmapLocationPopup(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import configparser
import glob
import multiprocessing
import os
import pickle
import platform
import re
import shutil
import webbrowser
from configparser import ConfigParser
from copy import deepcopy
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

import cv2
import numpy as np
import pandas as pd
import pkg_resources
import pyarrow as pa
from pyarrow import csv
from shapely.geometry import (LineString, MultiLineString, MultiPolygon, Point,
                              Polygon)

from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_if_dir_exists,
                                check_if_filepath_list_is_empty,
                                check_if_string_value_is_valid_video_timestamp,
                                check_instance, check_int,
                                check_nvidea_gpu_available, check_str,
                                check_that_column_exist)
from simba.utils.enums import ConfigKey, Dtypes, Formats, Keys
from simba.utils.errors import (DataHeaderError, DuplicationError,
                                FFMPEGCodecGPUError, FileExistError,
                                FrameRangeError, InvalidFilepathError,
                                InvalidFileTypeError, InvalidInputError,
                                InvalidVideoFileError,
                                MissingProjectConfigEntryError, NoDataError,
                                NoFilesFoundError, NotDirectoryError,
                                ParametersFileError, PermissionError)
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.warnings import (FileExistWarning, InvalidValueWarning,
                                  NoDataFoundWarning, NoFileFoundWarning)

PARSE_OPTIONS = csv.ParseOptions(delimiter=",")
READ_OPTIONS = csv.ReadOptions(encoding="utf8")


def read_df(
    file_path: Union[str, os.PathLike],
    file_type: Union[str, os.PathLike],
    has_index: Optional[bool] = True,
    remove_columns: Optional[List[str]] = None,
    usecols: Optional[List[str]] = None,
    anipose_data: Optional[bool] = False,
    check_multiindex: Optional[bool] = False,
    multi_index_headers_to_keep: Optional[int] = None,
) -> pd.DataFrame:
    """
    Read single tabular data file or pickle

    .. note::
       For improved runtime, defaults to :external:py:meth:`pyarrow.csv.write_cs` if file type is ``csv``.

    :parameter str file_path: Path to data file
    :parameter str file_type: Type of data. OPTIONS: 'parquet', 'csv', 'pickle'.
    :parameter Optional[bool]: If the input file has an initial index column. Default: True.
    :parameter Optional[List[str]] remove_columns: If not None, then remove columns in lits.
    :parameter Optional[List[str]] usecols: If not None, then keep columns in list.
    :parameter bool check_multiindex: check file is multi-index headers. Default: False.
    :parameter int multi_index_headers_to_keep: If reading multi-index file, and we want to keep one of the dropped multi-index levels as the header in the output file, specify the index of the multiindex hader as int.
    :return pd.DataFrame

    :example:
    >>> read_df(file_path='project_folder/csv/input_csv/Video_1.csv', file_type='csv', check_multiindex=True)
    """
    check_file_exist_and_readable(file_path=file_path)
    if file_type == Formats.CSV.value:
        try:
            df = csv.read_csv(
                file_path, parse_options=PARSE_OPTIONS, read_options=READ_OPTIONS
            )
            duplicate_headers = list(
                set([x for x in df.column_names if df.column_names.count(x) > 1])
            )
            if len(duplicate_headers) > 0:
                new_headers = [
                    duplicate_headers[0] + f"_{x}" for x in range(len(df.column_names))
                ]
                df = df.rename_columns(new_headers)
            if anipose_data:
                df = df.to_pandas()
                has_index = True
            else:
                df = df.to_pandas().iloc[:, 1:]
            if check_multiindex:
                header_col_cnt = get_number_of_header_columns_in_df(df=df)
                if multi_index_headers_to_keep is not None:
                    if multi_index_headers_to_keep not in list(
                        range(0, header_col_cnt)
                    ):
                        raise InvalidInputError(
                            msg=f"The selected multi-header index column {multi_index_headers_to_keep} does not exist in the multi-index header columns: {list(range(0, header_col_cnt))}",
                            source=read_df.__name__,
                        )
                    else:
                        new_header = list(
                            df.iloc[multi_index_headers_to_keep, :].values
                        )
                        new_header_xy = []
                        for header in list(set(new_header)):
                            new_header_xy.append(f"{header}_x")
                            new_header_xy.append(f"{header}_y"), new_header_xy.append(
                                f"{header}_likelihood"
                            )
                        df = df.drop(df.index[list(range(0, header_col_cnt))]).apply(
                            pd.to_numeric
                        )
                        df.columns = new_header_xy
                else:
                    df = df.drop(df.index[list(range(0, header_col_cnt))]).apply(
                        pd.to_numeric
                    )
            if not has_index:
                df = df.reset_index()
            else:
                df = df.reset_index(drop=True)
            df = df.astype(np.float32)

        except Exception as e:
            print(e, e.args)
            raise InvalidFileTypeError(
                msg=f"{file_path} is not a valid CSV file", source=read_df.__name__
            )
        if remove_columns:
            df = df[df.columns[~df.columns.isin(remove_columns)]]
        if usecols:
            df = df[df.columns[df.columns.isin(usecols)]]
    elif file_type == Formats.PARQUET.value:
        df = pd.read_parquet(file_path)
        if check_multiindex:
            header_col_cnt = get_number_of_header_columns_in_df(df=df)
            df = (
                df.drop(df.index[list(range(0, header_col_cnt))])
                .apply(pd.to_numeric)
                .reset_index(drop=True)
            )
        df = df.astype(np.float32)

    elif file_type == Formats.PICKLE.value:
        with open(file_path, "rb") as fp:
            df = pickle.load(fp)
    else:
        raise InvalidFileTypeError(
            msg=f"{file_type} is not a valid filetype OPTIONS: [pickle, csv, parquet]",
            source=read_df.__name__,
        )

    return df


def write_df(
    df: pd.DataFrame,
    file_type: str,
    save_path: Union[str, os.PathLike],
    multi_idx_header: bool = False,
) -> None:
    """
    Write single tabular data file.

    .. note::
       For improved runtime, defaults to ``pyarrow.csv`` if file_type == ``csv``.

    :parameter pd.DataFrame df: Pandas dataframe to save to disk.
    :parameter str file_type: Type of data. OPTIONS: ``parquet``, ``csv``,  ``pickle``.
    :parameter str save_path: Location where to store the data.
    :parameter bool check_multiindex: check if input file is multi-index headers. Default: False.

    :example:
    >>> write_df(df=df, file_type='csv', save_path='project_folder/csv/input_csv/Video_1.csv')
    """

    if file_type == Formats.CSV.value:
        if not multi_idx_header:
            df = df.drop("scorer", axis=1, errors="ignore")
            idx = np.arange(len(df)).astype(str)
            df.insert(0, "", idx)
            df = pa.Table.from_pandas(df=df)
            if "__index_level_0__" in df.column_names:
                df = df.remove_column(df.column_names.index("__index_level_0__"))
            csv.write_csv(df, save_path)
        else:
            try:
                df = df.drop("scorer", axis=1, errors="ignore")
            except TypeError:
                pass
            df.to_csv(save_path)
    elif file_type == Formats.PARQUET.value:
        df.to_parquet(save_path)
    elif file_type == Formats.PICKLE.value:
        try:
            with open(save_path, "wb") as f:
                pickle.dump(df, f, protocol=pickle.HIGHEST_PROTOCOL)
        except Exception as e:
            print(e.args[0])
            raise InvalidFileTypeError(
                msg="Data could not be saved as a pickle.", source=write_df.__name__
            )
    else:
        raise InvalidFileTypeError(
            msg=f"{file_type} is not a valid filetype OPTIONS: [csv, pickle, parquet]",
            source=write_df.__name__,
        )


def get_fn_ext(filepath: Union[os.PathLike, str]) -> (str, str, str):
    """
    Split file path into three components: (i) directory, (ii) file name, and (iii) file extension.

    :parameter str filepath: Path to file.
    :return str: File directory name
    :return str: File name
    :return str: File extension

    :example:
    >>> get_fn_ext(filepath='C:/My_videos/MyVideo.mp4')
    >>> ('My_videos', 'MyVideo', '.mp4')
    """
    file_extension = Path(filepath).suffix
    try:
        file_name = os.path.basename(filepath.rsplit(file_extension, 1)[0])
    except ValueError:
        raise InvalidFilepathError(
            msg=f"{filepath} is not a valid filepath", source=get_fn_ext.__name__
        )
    dir_name = os.path.dirname(filepath)
    return dir_name, file_name, file_extension


def read_config_entry(
    config: configparser.ConfigParser,
    section: str,
    option: str,
    data_type: str,
    default_value: Optional[Any] = None,
    options: Optional[List] = None,
) -> Union[float, int, str]:
    """
    Helper to read entry in SimBA project_config.ini parsed by configparser.ConfigParser.

    :param configparser.ConfigParser config: Parsed SimBA project_config.ini. Use :meth:`simba.utils.read_config_file` to parse file.
    :param str section: Section name of entry to parse.
    :param str option: Option name of entry to parse.
    :param str data_type: Type of data to parse. E.g., `str`, `int`, `float`.
    :param Optional[Any] default_value: If no matching entry can be found in the project_config.ini, use this as default.
    :param Optional[List] or None options: List of valid options. If not None, checks that the returned entry value exists in this list.
    :return Any

    :example:
    >>> read_config_entry(config='project_folder/project_config.ini', section='General settings', option='project_name', data_type='str')
    >>> 'two_animals_14_bps'
    """

    try:
        if config.has_option(section, option):
            if data_type == Dtypes.FLOAT.value:
                value = config.getfloat(section, option)
            elif data_type == Dtypes.INT.value:
                value = config.getint(section, option)
            elif data_type == Dtypes.STR.value:
                value = config.get(section, option).strip()
            elif data_type == Dtypes.FOLDER.value:
                value = config.get(section, option).strip()
                if not os.path.isdir(value):
                    raise NotDirectoryError(
                        msg=f"The SimBA config file includes paths to a folder ({value}) that does not exist.",
                        source=read_config_entry.__name__,
                    )
            if options != None:
                if value not in options:
                    raise InvalidInputError(
                        msg=f"{option} is set to {str(value)} in SimBA, but this is not among the valid options: ({options})",
                        source=read_config_entry.__name__,
                    )
                else:
                    return value
            return value

        elif default_value != None:
            return default_value
        else:
            raise MissingProjectConfigEntryError(
                msg=f"SimBA could not find an entry for option {option} under section {section} in the project_config.ini. Please specify the settings in the settings menu and make sure the path to your project config is correct",
                source=read_config_entry.__name__,
            )
    except ValueError:
        if default_value != None:
            return default_value
        else:
            raise MissingProjectConfigEntryError(
                msg=f"SimBA could not find an entry for option {option} under section {section} in the project_config.ini. Please specify the settings in the settings menu.",
                source=read_config_entry.__name__,
            )


def read_project_path_and_file_type(config: configparser.ConfigParser) -> (str, str):
    """
    Helper to read the path and file type of the SimBA project from the project_config.ini.

    :param configparser.ConfigParser config: parsed SimBA config in configparser.ConfigParser format
    :return str: The path of the project ``project_folder``.
    :return str: The set file type of the project (i.e., ``csv`` or ``parquet``).
    """

    project_path = read_config_entry(
        config=config,
        section=ConfigKey.GENERAL_SETTINGS.value,
        option=ConfigKey.PROJECT_PATH.value,
        data_type=ConfigKey.FOLDER_PATH.value,
    ).strip()
    file_type = read_config_entry(
        config=config,
        section=ConfigKey.GENERAL_SETTINGS.value,
        option=ConfigKey.FILE_TYPE.value,
        data_type=Dtypes.STR.value,
        default_value=Formats.CSV.value,
    ).strip()
    if not os.path.isdir(project_path):
        raise NotDirectoryError(
            msg=f"The project config file {config} has project path {project_path} that does not exist",
            source=read_project_path_and_file_type.__name__,
        )

    return project_path, file_type


def read_video_info_csv(file_path: Union[str, os.PathLike]) -> pd.DataFrame:
    """
    Read the project_folder/logs/video_info.csv of the SimBA project as a pd.DataFrame

    :parameter str file_path: Path to the SimBA project ``project_folder/logs/video_info.csv`` file
    :return pd.DataFrame
    :raise ParametersFileError: Invalid format of ``project_folder/logs/video_info.csv``.
    :raise InvalidValueWarning: Some videos are registered with FPS >= 1.

    :example:
    >>> read_video_info_csv(file_path='project_folder/logs/video_info.csv')
    """

    check_file_exist_and_readable(file_path=file_path)
    info_df = pd.read_csv(file_path)
    for c in [
        "Video",
        "fps",
        "Resolution_width",
        "Resolution_height",
        "Distance_in_mm",
        "pixels/mm",
    ]:
        if c not in info_df.columns:
            raise ParametersFileError(
                msg=f'The project "project_folder/logs/video_info.csv" does not not have an anticipated header ({c}). Please re-create the file and make sure each video has a {c} value',
                source=read_video_info_csv.__name__,
            )
    info_df["Video"] = info_df["Video"].astype(str)
    for c in [
        "fps",
        "Resolution_width",
        "Resolution_height",
        "Distance_in_mm",
        "pixels/mm",
    ]:
        try:
            info_df[c] = info_df[c].astype(float)
        except:
            raise ParametersFileError(
                msg=f'One or more values in the {c} column of the "project_folder/logs/video_info.csv" file could not be interpreted as a numeric value. Please re-create the file and make sure the entries in the {c} column are all numeric.',
                source=read_video_info_csv.__name__,
            )
    if info_df["fps"].min() <= 1:
        InvalidValueWarning(
            msg="Videos in your SimBA project have an FPS of 1 or less. Please use videos with more than one frame per second, or correct the inaccurate fps inside the `project_folder/logs/videos_info.csv` file",
            source=read_video_info_csv.__name__,
        )
    return info_df


def read_config_file(config_path: Union[str, os.PathLike]) -> configparser.ConfigParser:
    """
    Helper to parse SimBA project project_config.ini file

    :parameter str config_path: Path to project_config.ini file
    :return configparser.ConfigParser: parsed project_config.ini file
    :raise MissingProjectConfigEntryError: Invalid file format.

    :example:
    >>> read_config_file(config_path='project_folder/project_config.ini')
    """

    config = ConfigParser()
    try:
        config.read(config_path)
    except Exception as e:
        print(e.args)
        raise MissingProjectConfigEntryError(
            msg=f"{config_path} is not a valid project_config file. Please check the project_config.ini path.",
            source=read_config_entry.__name__,
        )
    return config


def get_video_meta_data(
    video_path: Union[str, os.PathLike], fps_as_int: bool = True
) -> dict:
    """
    Read video metadata (fps, resolution, frame cnt etc.) from video file (e.g., mp4).

    :parameter str video_path: Path to a video file.
    :parameter bool fps_as_int: If True, force video fps to int through floor rounding, else float. Default = True.
    :return dict: Video file meta data.

    :example:
    >>> get_video_meta_data('test_data/video_tests/Video_1.avi')
    {'video_name': 'Video_1', 'fps': 30, 'width': 400, 'height': 600, 'frame_count': 300, 'resolution_str': '400 x 600', 'video_length_s': 10}

    """

    video_data = {}
    cap = cv2.VideoCapture(video_path)
    _, video_data["video_name"], _ = get_fn_ext(video_path)
    video_data["fps"] = cap.get(cv2.CAP_PROP_FPS)
    if fps_as_int:
        video_data["fps"] = int(video_data["fps"])
    video_data["width"] = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    video_data["height"] = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    video_data["frame_count"] = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    for k, v in video_data.items():
        if v == 0:
            raise InvalidVideoFileError(
                msg=f'Video {video_data["video_name"]} either does not exist or has {k} of {str(v)} (full error video path: {video_path}).',
                source=get_video_meta_data.__name__,
            )
    video_data["resolution_str"] = str(
        f'{video_data["width"]} x {video_data["height"]}'
    )
    video_data["video_length_s"] = int(video_data["frame_count"] / video_data["fps"])
    return video_data


def remove_a_folder(folder_dir: Union[str, os.PathLike]) -> None:
    """Helper to remove a directory"""
    shutil.rmtree(folder_dir, ignore_errors=True)


def concatenate_videos_in_folder(
    in_folder: Union[str, os.PathLike],
    save_path: Union[str, os.PathLike],
    file_paths: Optional[List[Union[str, os.PathLike]]] = None,
    video_format: Optional[str] = "mp4",
    substring: Optional[str] = None,
    remove_splits: Optional[bool] = True,
    gpu: Optional[bool] = False,
) -> None:
    """
    Concatenate (temporally) all video files in a folder into a single video.

    .. important::
       Input video parts will be joined in alphanumeric order, should ideally have to have sequential numerical ordered file names, e.g., ``1.mp4``, ``2.mp4``....

    .. note::
       If substring and file_paths are both not None, then file_paths with be sliced and only file paths with substring will be retained.

    :parameter Union[str, os.PathLike] in_folder: Path to folder holding un-concatenated video files.
    :parameter Union[str, os.PathLike] save_path: Path to the saved the output file. Note: If the path exist, it will be overwritten
    :parameter Optional[List[Union[str, os.PathLike]]] file_paths: If not None, then the files that should be joined. If None, then all files. Default None.
    :parameter Optional[str] video_format: The format of the video clips that should be concatenated. Default: mp4.
    :parameter Optional[str] substring: If a string, then only videos in in_folder with a filename that contains substring will be joined. If None, then all are joined. Default: None.
    :parameter Optional[str] video_format: Format of the input video files in ``in_folder``. Default: ``mp4``.
    :parameter Optional[bool] remove_splits: If true, the input splits in the ``in_folder`` will be removed following concatenation. Default: True.
    """

    if not check_nvidea_gpu_available() and gpu:
        raise FFMPEGCodecGPUError(
            msg="No FFMpeg GPU codec found.",
            source=concatenate_videos_in_folder.__name__,
        )
    timer = SimbaTimer(start=True)
    if file_paths is None:
        files = glob.glob(in_folder + "/*.{}".format(video_format))
    else:
        for file_path in file_paths:
            check_file_exist_and_readable(file_path=file_path)
        files = file_paths
    check_if_filepath_list_is_empty(
        filepaths=files,
        error_msg=f"SIMBA ERROR: Cannot join videos in directory {in_folder}. The directory contain ZERO files in format {video_format}",
    )
    if substring is not None:
        sliced_paths = []
        for file_path in files:
            if substring in get_fn_ext(filepath=file_path)[1]:
                sliced_paths.append(file_path)
        check_if_filepath_list_is_empty(
            filepaths=sliced_paths,
            error_msg=f"SIMBA ERROR: Cannot join videos in directory {in_folder}. The directory contain ZERO files in format {video_format} with substring {substring}",
        )
        files = sliced_paths
    files.sort(key=lambda f: int(re.sub("\D", "", f)))
    temp_txt_path = Path(in_folder, "files.txt")
    if os.path.isfile(temp_txt_path):
        os.remove(temp_txt_path)
    with open(temp_txt_path, "w") as f:
        for file in files:
            f.write("file '" + str(Path(file)) + "'\n")

    if os.path.exists(save_path):
        os.remove(save_path)
    if check_nvidea_gpu_available() and gpu:
        returned = os.system(
            f'ffmpeg -hwaccel auto -c:v h264_cuvid -f concat -safe 0 -i "{temp_txt_path}" -c copy -hide_banner -loglevel info "{save_path}"'
        )
    else:
        returned = os.system(
            f'ffmpeg -f concat -safe 0 -i "{temp_txt_path}" "{save_path}" -c copy -hide_banner -loglevel info'
        )
    while True:
        if returned != 0:
            pass
        else:
            if remove_splits:
                print(in_folder, Path(in_folder).parents[0])
                remove_a_folder(folder_dir=Path(in_folder).parents[0])
            break
    timer.stop_timer()
    stdout_success(
        msg="Video concatenated",
        elapsed_time=timer.elapsed_time_str,
        source=concatenate_videos_in_folder.__name__,
    )


def get_bp_headers(body_parts_lst: List[str]) -> list:
    """
    Helper to create ordered list of all column header fields from body-part names for SimBA project dataframes.

    :parameter List[str] body_parts_lst: Body-part names in the SimBA prject
    :return List[str]: Body-part headers

    :examaple:
    >>> get_bp_headers(body_parts_lst=['Nose'])
    >>> ['Nose_x', 'Nose_y', 'Nose_p']
    """

    bp_headers = []
    for bp in body_parts_lst:
        c1, c2, c3 = (f"{bp}_x", f"{bp}_y", f"{bp}_p")
        bp_headers.extend((c1, c2, c3))
    return bp_headers


def read_video_info(
    vid_info_df: pd.DataFrame, video_name: str
) -> (pd.DataFrame, float, float):
    """
    Helper to read the metadata (pixels per mm, resolution, fps etc) from the video_info.csv for a single input file/video

    :parameter pd.DataFrame vid_info_df: Parsed ``project_folder/logs/video_info.csv`` file. This file can be parsed by :meth:`simba.utils.read_write.read_video_info_csv`.
    :parameter str video_name: Name of the video as represented in the ``Video`` column of the ``project_folder/logs/video_info.csv`` file.
    :returns pd.DataFrame: One row DataFrame representing the video in the ``project_folder/logs/video_info.csv`` file.
    :return float: The frame rate of the video as represented in the ``project_folder/logs/video_info.csv`` file
    :return float: The pixels per millimeter of the video as represented in the ``project_folder/logs/video_info.csv`` file
    :raise ParametersFileError: The video is not accurately represented in the ``project_folder/logs/video_info.csv`` file.

    :example:
    >>> video_info_df = read_video_info_csv(file_path='project_folder/logs/video_info.csv')
    >>> read_video_info(vid_info_df=vid_info_df, video_name='Together_1')
    """

    video_settings = vid_info_df.loc[vid_info_df["Video"] == video_name]
    if len(video_settings) > 1:
        raise DuplicationError(
            msg=f"SimBA found multiple rows in the project_folder/logs/video_info.csv named {str(video_name)}. Please make sure that each video name is represented ONCE in the video_info.csv",
            source=read_video_info.__name__,
        )
    elif len(video_settings) < 1:
        raise ParametersFileError(
            msg=f" SimBA could not find {str(video_name)} in the video_info.csv file. Make sure all videos analyzed are represented in the project_folder/logs/video_info.csv file.",
            source=read_video_info.__name__,
        )
    else:
        try:
            px_per_mm = float(video_settings["pixels/mm"])
            fps = float(video_settings["fps"])
            return video_settings, px_per_mm, fps
        except TypeError:
            raise ParametersFileError(
                msg=f"Make sure the videos that are going to be analyzed are represented with APPROPRIATE VALUES inside the project_folder/logs/video_info.csv file in your SimBA project. Could not interpret the fps, pixels per millimeter and/or fps as numerical values for video {video_name}",
                source=read_video_info.__name__,
            )


def find_all_videos_in_directory(
    directory: Union[str, os.PathLike],
    as_dict: Optional[bool] = False,
    raise_error: bool = False,
    video_formats: Optional[Tuple[str]] = (".avi", ".mp4", ".mov", ".flv", ".m4v"),
) -> Union[dict, list]:
    """
    Get all video file paths within a directory

    :param str directory: Directory to search for video files.
    :param bool as_dict: If True, returns dictionary with the video name as key and file path as value.
    :param bool raise_error: If True, raise error if no videos are found. Else, NoFileFoundWarning.
    :param Tuple[str] video_formats: Acceptable video formats. Default: '.avi', '.mp4', '.mov', '.flv', '.m4v'.
    :return List[str] or Dict[str, str]
    :raises NoFilesFoundError: If ``raise_error`` and ``directory`` has no files in formats ``video_formats``.

    :examples:
    >>> find_all_videos_in_directory(directory='project_folder/videos')
    """

    video_lst = []
    for i in os.listdir(directory):
        if i.lower().endswith(video_formats):
            video_lst.append(i)
    if not video_lst:
        if raise_error:
            raise NoFilesFoundError(
                f"No videos found in directory {directory} in formats {video_formats}."
            )
        video_lst.append("No videos found")
        NoFileFoundWarning(
            msg=f"No videos found in directory ({directory})",
            source=find_all_videos_in_directory.__name__,
        )

    if video_lst and as_dict:
        video_dict = {}
        for video_name in video_lst:
            video_path = os.path.join(directory, video_name)
            _, name, _ = get_fn_ext(filepath=video_path)
            video_dict[name] = video_path
        return video_dict

    return video_lst


def read_frm_of_video(
    video_path: Union[str, os.PathLike, cv2.VideoCapture],
    frame_index: int = 0,
    opacity: Optional[float] = None,
    size: Optional[Tuple[int, int]] = None,
    greyscale: Optional[bool] = False,
) -> np.ndarray:
    """
    Reads single image from video file.

    :param Union[str, os.PathLike] video_path: Path to video file, or cv2.VideoCapture object.
    :param int frame_index: The frame of video to return. Default: 1.
    :param Optional[int] opacity: Value between 0 and 100 or None. If float, returns image with opacity. 100 fully opaque. 0.0 fully transparant.
    :param Optional[Tuple[int, int]] size: If tuple, resizes the image to size. Else, returns original image size.
    :return np.ndarray: Image as numpy array.
    """

    check_instance(
        source=read_frm_of_video.__name__,
        instance=video_path,
        accepted_types=(str, cv2.VideoCapture),
    )
    if type(video_path) == str:
        check_file_exist_and_readable(file_path=video_path)
        video_meta_data = get_video_meta_data(video_path=video_path)
    else:
        video_meta_data = {"frame_count": int(video_path.get(cv2.CAP_PROP_FRAME_COUNT))}
    if (frame_index > video_meta_data["frame_count"]) or (frame_index < 0):
        raise FrameRangeError(
            msg=f'Frame {frame_index} is out of range: The video {video_path} contains {video_meta_data["frame_count"]} frames.',
            source=read_frm_of_video.__name__,
        )
    if type(video_path) == str:
        capture = cv2.VideoCapture(video_path)
    else:
        capture = video_path
    capture.set(cv2.CAP_PROP_POS_FRAMES, frame_index)
    ret, img = capture.read()
    if ret:
        if opacity:
            opacity = float(opacity / 100)
            check_float(
                name="Opacity",
                value=opacity,
                min_value=0.00,
                max_value=1.00,
                raise_error=True,
            )
            opacity = 1 - opacity
            h, w, clr = img.shape[:3]
            opacity_image = np.ones((h, w, clr), dtype=np.uint8) * int(255 * opacity)
            img = cv2.addWeighted(
                img.astype(np.uint8),
                1 - opacity,
                opacity_image.astype(np.uint8),
                opacity,
                0,
            )
        if size:
            img = cv2.resize(img, size, interpolation=cv2.INTER_LINEAR)
        if greyscale:
            if len(img.shape) > 2:
                img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    else:
        NoDataFoundWarning(
            msg=f"Frame {frame_index} for video {video_path} could not be read."
        )

    return img


def find_video_of_file(
    video_dir: Union[str, os.PathLike], filename: str, raise_error: bool = False
) -> Union[str, os.PathLike]:
    """
    Helper to find the video file with the SimBA project that represents a known data file path.

    :param str video_dir: Directory holding putative video file.
    :param str filename: Data file name, e.g., ``Video_1``.
    :param bool raise_error: If True, raise error if no file can be found. Else, print warning. Default: False
    :return str: Video path.
    :raise NoFilesFoundError: No video file representing file found.

    :examples:
    >>> find_video_of_file(video_dir='project_folder/videos', filename='Together_1')
    >>> 'project_folder/videos/Together_1.avi'

    """
    try:
        all_files_in_video_folder = [
            f for f in next(os.walk(video_dir))[2] if not f[0] == "."
        ]
    except StopIteration:
        raise NoFilesFoundError(
            msg=f"No files found in the {video_dir} directory",
            source=find_video_of_file.__name__,
        )
    all_files_in_video_folder = [
        os.path.join(video_dir, x) for x in all_files_in_video_folder
    ]
    return_path = None
    for file_path in all_files_in_video_folder:
        _, video_filename, ext = get_fn_ext(file_path)
        if (video_filename == filename) and (
            (ext.lower() == ".mp4") or (ext.lower() == ".avi")
        ):
            return_path = file_path

    if return_path is None and raise_error:
        raise NoFilesFoundError(
            msg=f"SimBA could not find a video file representing {filename} in the project video directory {video_dir}",
            source=find_video_of_file.__name__,
        )
    elif return_path is None:
        NoFileFoundWarning(
            msg=f"SimBA could not find a video file representing {filename} in the project video directory {video_dir}",
            source=find_video_of_file.__name__,
        )
    return return_path


def find_files_of_filetypes_in_directory(
    directory: str,
    extensions: list,
    raise_warning: Optional[bool] = True,
    raise_error: Optional[bool] = False,
) -> List[str]:
    """
    Find all files in a directory of specified extensions/types.

    :param str directory: Directory holding files.
    :param List[str] extensions: Accepted file extensions.
    :param bool raise_warning: If True, raise error if no files are found.

    :return List[str]: All files in ``directory`` with extensions.

    :example:
    >>> find_files_of_filetypes_in_directory(directory='project_folder/videos', extensions=['mp4', 'avi', 'png'], raise_warning=False)
    """

    try:
        all_files_in_folder = [
            f for f in next(os.walk(directory))[2] if not f[0] == "."
        ]
    except StopIteration:
        if raise_warning:
            raise NoFilesFoundError(
                msg=f"No files found in the {directory} directory with accepted extensions {str(extensions)}",
                source=find_files_of_filetypes_in_directory.__name__,
            )
        else:
            all_files_in_folder = []
            pass
    all_files_in_folder = [os.path.join(directory, x) for x in all_files_in_folder]
    accepted_file_paths = []
    for file_path in all_files_in_folder:
        _, file_name, ext = get_fn_ext(file_path)
        if ext.lower() in extensions:
            accepted_file_paths.append(file_path)
    if not accepted_file_paths and raise_warning:
        NoFileFoundWarning(
            msg=f"SimBA could not find any files with accepted extensions {extensions} in the {directory} directory",
            source=find_files_of_filetypes_in_directory.__name__,
        )
    if not accepted_file_paths and raise_error:
        raise NoDataError(
            msg=f"SimBA could not find any files with accepted extensions {extensions} in the {directory} directory",
            source=find_files_of_filetypes_in_directory.__name__,
        )
    return accepted_file_paths


def convert_parquet_to_csv(directory: str) -> None:
    """
    Convert all parquet files in a directory to csv format.

    :param str directory: Path to directory holding parquet files
    :raise NoFilesFoundError: The directory has no ``parquet`` files.

    :examples:
    >>> convert_parquet_to_csv(directory='project_folder/csv/input_csv')
    """

    if not os.path.isdir(directory):
        raise NotDirectoryError(
            msg="SIMBA ERROR: {} is not a valid directory".format(directory),
            source=convert_parquet_to_csv.__name__,
        )
    files_found = glob.glob(directory + "/*.parquet")
    if len(files_found) < 1:
        raise NoFilesFoundError(
            "SIMBA ERROR: No parquet files (with .parquet file ending) found in the {} directory".format(
                directory
            ),
            source=convert_parquet_to_csv.__name__,
        )
    for file_cnt, file_path in enumerate(files_found):
        print("Reading in {} ...".format(os.path.basename(file_path)))
        df = pd.read_parquet(file_path)
        new_file_path = os.path.join(
            directory, os.path.basename(file_path).replace(".parquet", ".csv")
        )
        if "scorer" in df.columns:
            df = df.set_index("scorer")
        df.to_csv(new_file_path)
        print("Saved {}...".format(new_file_path))
    stdout_success(
        msg=f"{str(len(files_found))} parquet files in {directory} converted to csv",
        source=convert_parquet_to_csv.__name__,
    )


def convert_csv_to_parquet(directory: Union[str, os.PathLike]) -> None:
    """
    Convert all csv files in a folder to parquet format.

    :param str directory: Path to directory holding csv files.
    :raise NoFilesFoundError: The directory has no ``csv`` files.

    :examples:
    >>> convert_parquet_to_csv(directory='project_folder/csv/input_csv')
    """
    if not os.path.isdir(directory):
        raise NotDirectoryError(
            msg="SIMBA ERROR: {} is not a valid directory".format(directory),
            source=convert_csv_to_parquet.__name__,
        )
    files_found = glob.glob(directory + "/*.csv")
    if len(files_found) < 1:
        raise NoFilesFoundError(
            msg="SIMBA ERROR: No parquet files (with .csv file ending) found in the {} directory".format(
                directory
            ),
            source=convert_csv_to_parquet.__name__,
        )
    print("Converting {} files...".format(str(len(files_found))))
    for file_cnt, file_path in enumerate(files_found):
        print("Reading in {} ...".format(os.path.basename(file_path)))
        df = pd.read_csv(file_path)
        new_file_path = os.path.join(
            directory, os.path.basename(file_path).replace(".csv", ".parquet")
        )
        df.to_parquet(new_file_path)
        print("Saved {}...".format(new_file_path))
    stdout_success(
        msg=f"{str(len(files_found))} csv files in {directory} converted to parquet",
        source=convert_csv_to_parquet.__name__,
    )


def get_file_name_info_in_directory(
    directory: Union[str, os.PathLike], file_type: str
) -> Dict[str, str]:
    """
    Get dict of all file paths in a directory with specified extension as values and file base names as keys.

    :param str directory: Directory containing files.
    :param str file_type: File-type in ``directory`` of interest
    :return dict: All found files as values and file base names as keys.

    :example:
    >>> get_file_name_info_in_directory(directory='C:\project_folder\csv\machine_results', file_type='csv')
    >>> {'Video_1': 'C:\project_folder\csv\machine_results\Video_1'}
    """

    results = {}
    file_paths = glob.glob(directory + "/*." + file_type)
    for file_path in file_paths:
        _, file_name, ext = get_fn_ext(file_path)
        results[file_name] = file_path

    return results


def archive_processed_files(
    config_path: Union[str, os.PathLike], archive_name: str
) -> None:
    """
    Archive files within a SimBA project.

    :param str config_path: Path to SimBA project ``project_config.ini``.
    :param str archive_name: Name of archive.

    .. seealso::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario4_new.md>`_

    :example:
    >>> archive_processed_files(config_path='project_folder/project_config.ini', archive_name='my_archive')
    """

    config = read_config_file(config_path=config_path)
    file_type = read_config_entry(
        config,
        ConfigKey.GENERAL_SETTINGS.value,
        ConfigKey.FILE_TYPE.value,
        "str",
        "csv",
    )
    project_path = read_config_entry(
        config,
        ConfigKey.GENERAL_SETTINGS.value,
        ConfigKey.PROJECT_PATH.value,
        data_type=ConfigKey.FOLDER_PATH.value,
    )
    videos_dir = os.path.join(project_path, "videos")
    csv_dir = os.path.join(os.path.dirname(config_path), "csv")
    log_path = os.path.join(project_path, "logs")
    video_info_path = os.path.join(log_path, "video_info.csv")
    csv_subdirs, file_lst = [], []
    for content_name in os.listdir(csv_dir):
        if os.path.isdir(os.path.join(csv_dir, content_name)):
            csv_subdirs.append(os.path.join(csv_dir, content_name))

    for subdirectory in csv_subdirs:
        subdirectory_files = [
            x for x in glob.glob(subdirectory + "/*") if os.path.isfile(x)
        ]
        for file_path in subdirectory_files:
            directory, file_name, ext = get_fn_ext(
                os.path.join(subdirectory, file_path)
            )
            if ext == ".{}".format(file_type):
                file_lst.append(os.path.join(subdirectory, file_path))

    if len(file_lst) < 1:
        raise NoFilesFoundError(
            msg="SIMBA ERROR: No data files located in your project_folder/csv sub-directories in the worflow file format {}".format(
                file_type
            ),
            source=archive_processed_files.__name__,
        )

    for file_path in file_lst:
        file_folder = os.path.dirname(file_path)
        save_directory = os.path.join(file_folder, archive_name)
        save_file_path = os.path.join(save_directory, os.path.basename(file_path))
        if not os.path.exists(save_directory):
            os.mkdir(save_directory)
        print("Moving file {}...".format(file_path))
        shutil.move(file_path, save_file_path)

    log_archive_path = os.path.join(log_path, archive_name)
    if not os.path.exists(log_archive_path):
        os.mkdir(log_archive_path)
    if os.path.isfile(video_info_path):
        save_file_path = os.path.join(log_archive_path, "video_info.csv")
        print("Moving file {}...".format(video_info_path))
        shutil.move(video_info_path, save_file_path)

    videos_file_paths = [f for f in glob.glob(videos_dir) if os.path.isfile(f)]
    video_archive_path = os.path.join(videos_dir, archive_name)
    if not os.path.exists(video_archive_path):
        os.mkdir(video_archive_path)
    for video_file in videos_file_paths:
        save_video_path = os.path.join(video_archive_path, os.path.basename(video_file))
        shutil.move(video_file, save_video_path)
    stdout_success(msg="Archiving completed", source=archive_processed_files.__name__)


def str_2_bool(input_str: str) -> bool:
    """
    Helper to convert string representation of bool to bool.

    :example:
    >>> str_2_bool(input_str='yes')
    >>> True
    """
    return input_str.lower() in ("yes", "true", "1")


def tabulate_clf_info(clf_path: Union[str, os.PathLike]) -> None:
    """
    Print the hyperparameters and creation date of a pickled classifier.

    :param str clf_path: Path to classifier
    :raise InvalidFilepathError: The file is not a pickle or not a scikit-learn RF classifier.
    """

    _, clf_name, _ = get_fn_ext(clf_path)
    check_file_exist_and_readable(file_path=clf_path)
    try:
        clf_obj = pickle.load(open(clf_path, "rb"))
    except:
        raise InvalidFilepathError(
            msg=f"The {clf_path} file is not a pickle file",
            source=tabulate_clf_info.__name__,
        )
    try:
        clf_features_no = clf_obj.n_features_
        clf_criterion = clf_obj.criterion
        clf_estimators = clf_obj.n_estimators
        clf_min_samples_leaf = clf_obj.min_samples_split
        clf_n_jobs = clf_obj.n_jobs
        clf_verbose = clf_obj.verbose
        if clf_verbose == 1:
            clf_verbose = True
        if clf_verbose == 0:
            clf_verbose = False
    except:
        raise InvalidFilepathError(
            msg=f"The {clf_path} file is not an scikit-learn RF classifier",
            source=tabulate_clf_info.__name__,
        )
    creation_time = "Unknown"
    try:
        if platform.system() == "Windows":
            creation_time = os.path.getctime(clf_path)
        elif platform.system() == "Darwin":
            creation_time = os.stat(clf_path)
            creation_time = creation_time.st_birthtime
    except AttributeError:
        pass
    if creation_time != "Unknown":
        creation_time = str(
            datetime.utcfromtimestamp(creation_time).strftime("%Y-%m-%d %H:%M:%S")
        )

    print(str(clf_name), "CLASSIFIER INFORMATION")
    for name, val in zip(
        [
            "NUMBER OF FEATURES",
            "NUMBER OF TREES",
            "CLASSIFIER CRITERION",
            "CLASSIFIER_MIN_SAMPLE_LEAF",
            "CLASSIFIER_N_JOBS",
            "CLASSIFIER VERBOSE SETTING",
            "CLASSIFIER PATH",
            "CLASSIFIER CREATION TIME",
        ],
        [
            clf_features_no,
            clf_estimators,
            clf_criterion,
            clf_min_samples_leaf,
            clf_n_jobs,
            clf_verbose,
            clf_path,
            str(creation_time),
        ],
    ):
        print(name + ": " + str(val))


def get_all_clf_names(config: configparser.ConfigParser, target_cnt: int) -> List[str]:
    """
    Get all classifier names in a SimBA project.

    :param configparser.ConfigParser config: Parsed SimBA project_config.ini
    :param int target_cnt: Count of models in SimBA project
    :return List[str]: Classifier model names

    :example:
    >>> get_all_clf_names(config=config, target_cnt=2)
    >>> ['Attack', 'Sniffing']
    """

    model_names = []
    for i in range(target_cnt):
        entry_name = "target_name_{}".format(str(i + 1))
        model_names.append(
            read_config_entry(
                config,
                ConfigKey.SML_SETTINGS.value,
                entry_name,
                data_type=Dtypes.STR.value,
            )
        )
    return model_names


def read_meta_file(meta_file_path: Union[str, os.PathLike]) -> dict:
    """
    Read in single SimBA modelconfig meta file CSV to python dictionary.

    :param str meta_file_path: Path to SimBA config meta file
    :return dict: Dictionary holding model parameters.

    :example:
    >>> read_meta_file('project_folder/configs/Attack_meta_0.csv')
    >>> {'Classifier_name': 'Attack', 'RF_n_estimators': 2000, 'RF_max_features': 'sqrt', 'RF_criterion': 'gini', ...}
    """
    check_file_exist_and_readable(file_path=meta_file_path)
    return pd.read_csv(meta_file_path, index_col=False).to_dict(orient="records")[0]


def read_simba_meta_files(folder_path: str, raise_error: bool = False) -> List[str]:
    """
    Read in paths of SimBA model config files directory (`project_folder/configs'). Consider files that have `meta` suffix only.

    :param str folder_path: directory with SimBA model config meta files
    :param bool raise_error: If True, raise error if no files are found with ``meta`` suffix. Else, print warning. Default: False.
    :return List[str]: List of paths to  SimBA model config meta files.

    :example:
    >>> read_simba_meta_files(folder_path='/project_folder/configs')
    >>> ['project_folder/configs/Attack_meta_1.csv', 'project_folder/configs/Attack_meta_0.csv']
    """

    file_paths = find_files_of_filetypes_in_directory(
        directory=folder_path, extensions=[".csv"]
    )
    meta_file_lst = []
    for i in file_paths:
        if i.__contains__("meta"):
            meta_file_lst.append(os.path.join(folder_path, i))
    if len(meta_file_lst) == 0 and not raise_error:
        NoFileFoundWarning(
            msg=f'The training meta-files folder in your project ({folder_path}) does not have any meta files inside it (no files in this folder has the "meta" substring in the filename)',
            source=read_simba_meta_files.__name__,
        )
    elif len(meta_file_lst) == 0 and raise_error:
        raise NoFilesFoundError(
            msg=f'The training meta-files folder in your project ({folder_path}) does not have any meta files inside it (no files in this folder has the "meta" substring in the filename)',
            source=read_simba_meta_files.__name__,
        )

    return meta_file_lst


def find_core_cnt() -> Tuple[int, int]:
    """
    Find the local cpu count and quarter of the cpu counts.

    :return int: The local cpu count
    :return int: The local cpu count // 4

    :example:
    >>> find_core_cnt()
    >>> (8, 2)

    """
    cpu_cnt = multiprocessing.cpu_count()
    cpu_cnt_to_use = int(cpu_cnt / 4)
    if cpu_cnt_to_use < 1:
        cpu_cnt_to_use = 1
    return cpu_cnt, cpu_cnt_to_use


def get_number_of_header_columns_in_df(df: pd.DataFrame) -> int:
    """
    Returns the count of non-numerical header rows in dataframe. E.g., can be helpful to determine if dataframe is multi-index columns.

    :param pd.DataFrame df: Dataframe to check the count of non-numerical header rows for.

    :example:
    >>> get_number_of_header_columns_in_df(df='project_folder/csv/input_csv/Video_1.csv')
    >>> 3
    """
    for i in range(len(df)):
        try:
            temp = df.iloc[i:].apply(pd.to_numeric).reset_index(drop=True)
            return i
        except ValueError:
            pass
    raise DataHeaderError(
        msg="Could find the count of header columns in dataframe, all rows appear non-numeric",
        source=get_number_of_header_columns_in_df.__name__,
    )


def get_memory_usage_of_df(df: pd.DataFrame) -> Dict[str, float]:
    """
    Get the RAM memory usage of a dataframe.

    :param pd.DataFrame df: Parsed dataframe
    :return dict: The memory usage of the dataframe in bytes, mb, and gb.

    :example:
    >>> df = pd.DataFrame(np.random.randint(0,100,size=(100, 4)), columns=list('ABCD'))
    >>> {'bytes': 3328, 'megabytes': 0.003328, 'gigabytes': 3e-06}
    """

    results = {}
    results["bytes"] = df.memory_usage(index=True).sum()
    results["megabytes"] = round(results["bytes"] / 1000000, 6)
    results["gigabytes"] = round(results["bytes"] / 1000000000, 6)
    return results


def copy_single_video_to_project(
    simba_ini_path: Union[str, os.PathLike],
    source_path: Union[str, os.PathLike],
    symlink: bool = False,
    allowed_video_formats: Optional[Tuple[str]] = ("avi", "mp4"),
    overwrite: Optional[bool] = False,
) -> None:
    """
    Import single video file to SimBA project

    :param Union[str, os.PathLike] simba_ini_path: path to SimBA project config file in Configparser format
    :param Union[str, os.PathLike] source_path: Path to video file outside SimBA project.
    :param Optional[bool] symlink: If True, creates soft copy rather than hard copy. Default: False.
    :param Optional[Tuple[str]] allowed_video_formats: Allowed video formats. DEFAULT: avi or mp4
    :param Optional[bool] overwrite: If True, overwrites existing video if it exists in SimBA project. Else, raise FileExistError.
    """

    timer = SimbaTimer(start=True)
    _, file_name, file_ext = get_fn_ext(source_path)
    check_file_exist_and_readable(file_path=source_path)
    print("Copying video {} file...".format(file_name))
    if file_ext[1:].lower().strip() not in allowed_video_formats:
        raise InvalidFileTypeError(
            msg="SimBA works best with avi and mp4 video-files. Or please convert your videos to mp4 or avi to continue before importing it.",
            source=copy_single_video_to_project.__name__,
        )
    new_filename = os.path.join(file_name + file_ext)
    destination = os.path.join(os.path.dirname(simba_ini_path), "videos", new_filename)
    if os.path.isfile(destination) and not overwrite:
        raise FileExistError(
            msg=f"{file_name} already exist in SimBA project. To import, delete this video file before importing the new video file with the same name.",
            source=copy_single_video_to_project.__name__,
        )
    else:
        if not symlink:
            shutil.copy(source_path, destination)
        else:
            try:
                if os.path.isfile(destination):
                    os.remove(destination)
                os.symlink(source_path, destination)
            except OSError as e:
                raise PermissionError(
                    msg="Symbolic link privilege not held. Try running SimBA in terminal opened in admin mode",
                    source=copy_single_video_to_project.__name__,
                )
        timer.stop_timer()
        if not symlink:
            stdout_success(
                msg=f"Video {file_name} imported to SimBA project (project_folder/videos directory",
                elapsed_time=timer.elapsed_time_str,
                source=copy_single_video_to_project.__name__,
            )
        else:
            stdout_success(
                msg=f"Video {file_name}  SYMLINK imported to SimBA project (project_folder/videos directory",
                elapsed_time=timer.elapsed_time_str,
                source=copy_single_video_to_project.__name__,
            )


def copy_multiple_videos_to_project(
    config_path: Union[str, os.PathLike],
    source: Union[str, os.PathLike],
    file_type: str,
    symlink: Optional[bool] = False,
    allowed_video_formats: Optional[Tuple[str]] = ("avi", "mp4"),
) -> None:
    """
    Import directory of videos to SimBA project.

    :param Union[str, os.PathLike] simba_ini_path: path to SimBA project config file in Configparser format
    :param Union[str, os.PathLike] source_path: Path to directory with video files outside SimBA project.
    :param str file_type: Video format of imported videos (i.e.,: mp4 or avi)
    :param Optional[bool] symlink: If True, creates soft copies rather than hard copies. Default: False.
    :param Optional[Tuple[str]] allowed_video_formats: Allowed video formats. DEFAULT: avi or mp4
    """

    multiple_video_timer = SimbaTimer(start=True)
    if file_type.lower().strip() not in allowed_video_formats:
        raise InvalidFileTypeError(
            msg="SimBA only works with avi and mp4 video files (Please enter mp4 or avi in entrybox). Or convert your videos to mp4 or avi to continue.",
            source=copy_multiple_videos_to_project.__name__,
        )
    video_path_lst = find_all_videos_in_directory(
        directory=source, video_formats=(file_type), raise_error=True
    )
    video_path_lst = [os.path.join(source, x) for x in video_path_lst]
    if len(video_path_lst) == 0:
        raise NoFilesFoundError(
            msg=f"SIMBA ERROR: No videos found in {source} directory of file-type {file_type}",
            source=copy_multiple_videos_to_project.__name__,
        )
    destination_dir = os.path.join(os.path.dirname(config_path), "videos")
    for file_cnt, file_path in enumerate(video_path_lst):
        timer = SimbaTimer(start=True)
        dir_name, filebasename, file_extension = get_fn_ext(file_path)
        file_extension = file_extension.lower()
        newFileName = os.path.join(filebasename + file_extension)
        dest1 = os.path.join(destination_dir, newFileName)
        if os.path.isfile(dest1):
            FileExistWarning(
                msg=f"{filebasename} already exist in SimBA project. Skipping video...",
                source=copy_multiple_videos_to_project.__name__,
            )
        else:
            if not symlink:
                shutil.copy(file_path, dest1)
            else:
                try:
                    os.symlink(file_path, dest1)
                except OSError:
                    raise PermissionError(
                        msg="Symbolic link privilege not held. Try running SimBA in terminal opened in admin mode"
                    )
            timer.stop_timer()
            if not symlink:
                print(
                    f"{filebasename} copied to project (Video {file_cnt + 1}/{len(video_path_lst)}, elapsed timer {timer.elapsed_time_str}s)..."
                )
            else:
                print(
                    f"{filebasename} copied to project (SYMLINK) (Video {file_cnt + 1}/{len(video_path_lst)}, elapsed timer {timer.elapsed_time_str}s)..."
                )

    multiple_video_timer.stop_timer()
    stdout_success(
        msg=f"{len(video_path_lst)} videos copied to project.",
        elapsed_time=multiple_video_timer.elapsed_time_str,
        source=copy_multiple_videos_to_project.__name__,
    )


def find_all_videos_in_project(
    videos_dir: Union[str, os.PathLike], basename: Optional[bool] = False
) -> List[str]:
    """
    Get filenames of .avi and .mp4 files within a directory

    :param str videos_dir: Directory holding video files.
    :param bool basename: If true returns basenames, else file paths.

    :example:
    >>> find_all_videos_in_project(videos_dir='project_folder/videos')
    >>> ['project_folder/videos/Together_2.avi', 'project_folder/videos/Together_3.avi', 'project_folder/videos/Together_1.avi']

    """
    video_paths = []
    file_paths_in_folder = [f for f in next(os.walk(videos_dir))[2] if not f[0] == "."]
    file_paths_in_folder = [os.path.join(videos_dir, f) for f in file_paths_in_folder]
    for file_cnt, file_path in enumerate(file_paths_in_folder):
        try:
            _, file_name, file_ext = get_fn_ext(file_path)
        except ValueError:
            raise InvalidFilepathError(
                msg=f"{file_path} is not a valid filepath",
                source=find_all_videos_in_project.__name__,
            )
        if (file_ext.lower() == ".mp4") or (file_ext.lower() == ".avi"):
            if not basename:
                video_paths.append(file_path)
            else:
                video_paths.append(file_name)
    if len(video_paths) == 0:
        raise NoFilesFoundError(
            msg=f"No videos in mp4 or avi format found imported to SimBA project in the {videos_dir} directory",
            source=find_all_videos_in_project.__name__,
        )
    else:
        return video_paths


def check_if_hhmmss_timestamp_is_valid_part_of_video(
    timestamp: str, video_path: Union[str, os.PathLike]
) -> None:
    """
    Helper to check that a timestamp in HH:MM:SS format is a valid timestamp in a video file.

    :param str timestamp: Timestamp in HH:MM:SS format.
    :param str video_path: Path to a video file.
    :raises FrameRangeError: If timestamp is not in the video file. E.g., timestamp 00:01:00 will raise FrameRangeError if the video is 59s long.

    :example:
    >>> check_if_hhmmss_timestamp_is_valid_part_of_video(timestamp='01:00:05', video_path='/Users/simon/Desktop/video_tests/Together_1.avi')
    >>> "FrameRangeError: The timestamp '01:00:05' does not occur in video Together_1.avi, the video has length 10s"
    """

    check_file_exist_and_readable(file_path=video_path)
    check_if_string_value_is_valid_video_timestamp(value=timestamp, name="Timestamp")
    video_meta_data = get_video_meta_data(video_path=video_path)
    h, m, s = timestamp.split(":")
    time_stamp_in_seconds = int(h) * 3600 + int(m) * 60 + int(s)
    if not video_meta_data["video_length_s"] >= time_stamp_in_seconds:
        video_length_str = timedelta(seconds=video_meta_data["video_length_s"])
        raise FrameRangeError(
            msg=f'The timestamp {timestamp} does not occur in video {video_meta_data["video_name"]}, the video has length {video_length_str}',
            source=check_if_hhmmss_timestamp_is_valid_part_of_video.__name__,
        )


def find_time_stamp_from_frame_numbers(
    start_frame: int, end_frame: int, fps: float
) -> List[str]:
    """
    Given start and end frame numbers and frames per second (fps), return a list of formatted time stamps
    corresponding to the frame range start and end time.

    :param int start_frame: The starting frame index.
    :param int end_frame: The ending frame index.
    :param float fps: Frames per second.
    :return List[str]: A list of time stamps in the format 'HH:MM:SS:MS'.

    :example:
    >>> find_time_stamp_from_frame_numbers(start_frame=11, end_frame=20, fps=3.4)
    >>> ['00:00:03:235', '00:00:05:882']
    """

    def get_time(frame_index, fps):
        total_seconds = frame_index / fps
        milliseconds = int((total_seconds % 1) * 1000)
        total_seconds = int(total_seconds)
        seconds = total_seconds % 60
        total_seconds //= 60
        minutes = total_seconds % 60
        hours = total_seconds // 60
        return "{:02d}:{:02d}:{:02d}.{:03d}".format(
            hours, minutes, seconds, milliseconds
        )

    check_int(name="start_frame", value=start_frame, min_value=0)
    check_int(name="end_frame", value=end_frame, min_value=0)
    check_float(name="FPS", value=fps, min_value=1)
    if start_frame > end_frame:
        raise FrameRangeError(
            msg=f"Start frame ({start_frame}) cannot be before end frame ({end_frame})"
        )

    return [get_time(start_frame, fps), get_time(end_frame, fps)]


def read_roi_data(
    roi_path: Union[str, os.PathLike]
) -> (pd.DataFrame, pd.DataFrame, pd.DataFrame):
    """
    Method to read in ROI definitions from SimBA project
    """
    check_file_exist_and_readable(file_path=roi_path)
    try:
        rectangles_df = pd.read_hdf(roi_path, key=Keys.ROI_RECTANGLES.value).dropna(
            how="any"
        )
        circles_df = pd.read_hdf(roi_path, key=Keys.ROI_CIRCLES.value).dropna(how="any")
        polygon_df = pd.read_hdf(roi_path, key=Keys.ROI_POLYGONS.value)
    except:
        raise InvalidFileTypeError(
            msg=f"{roi_path} is not a valid SimBA ROI definitions file",
            source=read_roi_data.__name__,
        )
    if "Center_XCenter_Y" in polygon_df.columns:
        polygon_df = polygon_df.drop(["Center_XCenter_Y"], axis=1)
    polygon_df = polygon_df.dropna(how="any")

    return rectangles_df, circles_df, polygon_df


def create_directory(path: Union[str, os.PathLike]):
    if not os.path.exists(path):
        os.makedirs(path)
    else:
        pass


def find_max_vertices_coordinates(
    shapes: List[Union[Polygon, LineString, MultiPolygon, Point]],
    buffer: Optional[int] = None,
) -> Tuple[int, int]:
    """
    Find the maximum x and y coordinates among the vertices of a list of Shapely geometries.

    Can be useful for plotting puposes, to dtermine the rquired size of the canvas to fit all geometries.

    :param List[Union[Polygon, LineString, MultiPolygon, Point]] shapes: A list of Shapely geometries including Polygons, LineStrings, MultiPolygons, and Points.
    :param Optional[int] buffer: If int, adds to maximum x and y.
    :returns Tuple[int, int]: A tuple containing the maximum x and y coordinates found among the vertices.

    :example:
    >>> polygon = Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
    >>> line = LineString([(1, 1), (2, 2), (3, 1), (4, 0)])
    >>> multi_polygon = MultiPolygon([Polygon([(0, 0), (1, 0), (1, 1), (0, 1)]), Polygon([(1, 1), (2, 1), (2, 2), (1, 2)])])
    >>> point = Point(3, 4)
    >>> find_max_vertices_coordinates([polygon, line, multi_polygon, point])
    >>> (4, 4)
    """

    for shape in shapes:
        check_instance(
            source=find_max_vertices_coordinates.__name__,
            instance=shape,
            accepted_types=(Polygon, LineString, MultiPolygon, Point, MultiLineString),
        )

    max_x, max_y = -np.inf, -np.inf
    for shape in shapes:
        if isinstance(shape, Polygon):
            for vertex in shape.exterior.coords:
                max_x, max_y = max(max_x, vertex[0]), max(max_y, vertex[1])

        if isinstance(shape, MultiPolygon):
            for polygon in shape.geoms:
                for vertex in polygon.exterior.coords:
                    max_x, max_y = max(max_x, vertex[0]), max(max_y, vertex[1])

        if isinstance(shape, LineString):
            for vertex in shape.coords:
                max_x, max_y = max(max_x, vertex[0]), max(max_y, vertex[1])

        if isinstance(shape, Point):
            max_x, max_y = max(max_x, shape.coords[0][0]), max(
                max_y, shape.coords[0][1]
            )

        if isinstance(shape, MultiLineString):
            for line in shape.geoms:
                for vertex in line.coords:
                    max_x, max_y = max(max_x, vertex[0]), max(max_y, vertex[1])

    if buffer:
        check_int(name="Buffer", value=buffer, min_value=1)
        max_x += buffer
        max_y += buffer

    return int(max_x), int(max_y)


def clean_sleap_file_name(filename: str) -> str:
    """
    Clean a SLEAP input filename by removing '.analysis' suffix and project name prefix to match orginal video name.

     .. note::
       Modified from `vtsai881 <https://github.com/vtsai881>`_.

    :param str filename: The original filename to be cleaned to match video name.
    :returns str: The cleaned filename.

    :example:
    >>> clean_sleap_file_name("projectname.v00x.00x_videoname.analysis.csv")
    >>> 'videoname.csv'
    >>> clean_sleap_file_name("projectname.v00x.00x_videoname.analysis.h5")
    >>> 'videoname.h5'
    """

    SLEAP_CSV_SUBSTR = ".analysis"
    if (SLEAP_CSV_SUBSTR in filename) and ("_" in filename):
        return filename.replace(filename.split("_")[0] + "_", "").replace(
            SLEAP_CSV_SUBSTR, ""
        )
    else:
        return filename


def clean_sleap_filenames_in_directory(dir: Union[str, os.PathLike]) -> None:
    """
    Clean up SLEAP input filenames in the specified directory by removing a prefix
    and a suffix, and renaming the files to match the names of the original video files.

    .. note::
       Modified from `vtsai881 <https://github.com/vtsai881>`_.

    :param Union[str, os.PathLike] dir: The directory path where the SLEAP CSV or H5 files are located.

    :example:
    >>> clean_sleap_filenames_in_directory(dir='/Users/simon/Desktop/envs/troubleshooting/Hornet_SLEAP/import/')
    """

    SLEAP_CSV_SUBSTR = ".analysis"
    check_if_dir_exists(in_dir=dir)
    for file_path in glob.glob(
        dir + f"/*.{Formats.CSV.value}" + f"/*.{Formats.H5.value}"
    ):
        file_name = os.path.basename(p=file_path)
        if (SLEAP_CSV_SUBSTR in file_name) and ("_" in file_name):
            new_name = os.path.join(
                dir,
                file_name.replace(file_name.split("_")[0] + "_", "").replace(
                    SLEAP_CSV_SUBSTR, ""
                ),
            )
            os.rename(file_path, new_name)
        else:
            pass


def copy_files_in_directory(
    in_dir: Union[str, os.PathLike],
    out_dir: Union[str, os.PathLike],
    raise_error: bool = True,
    filetype: Optional[str] = None,
) -> None:
    """
    Copy files from the specified input directory to the output directory.

    :param Union[str, os.PathLike] in_dir: The input directory from which files will be copied.
    :param Union[str, os.PathLike] out_dir: The output directory where files will be copied to.
    :param bool raise_error: If True, raise an error if no files are found in the input directory. Default is True.
    :param Optional[str] filetype: If specified, only copy files with the given file extension. Default is None, meaning all files will be copied.

    :example:
    >>> copy_files_in_directory('/input_dir', '/output_dir', raise_error=True, filetype='txt')
    """

    check_if_dir_exists(in_dir=in_dir)
    if not os.listdir(out_dir):
        os.makedirs(out_dir)
    if filetype is not None:
        file_paths = glob.glob(in_dir + f"/*.{filetype}")
    else:
        file_paths = glob.glob(in_dir + f"/*.")
    if len(file_paths) == 0 and raise_error:
        raise NoFilesFoundError(
            msg=f"No files found in {in_dir}", source=copy_files_in_directory.__name__
        )
    elif len(file_paths) == 0:
        pass
    else:
        for file_path in file_paths:
            shutil.copy(file_path, out_dir)


def remove_files(
    file_paths: List[Union[str, os.PathLike]], raise_error: Optional[bool] = False
) -> None:
    """
    Delete (remove) the files specified within a list of filepaths.

    :param Union[str, os.PathLike] file_paths: A list of file paths to be removed.
    :param Optional[bool] raise_error: If True, raise exceptions for errors during file deletion. Else, pass. Defaults to False.

    :examples:
    >>> file_paths = ['/path/to/file1.txt', '/path/to/file2.txt']
    >>> remove_files(file_paths, raise_error=True)
    """

    for file_path in file_paths:
        if not os.path.isfile(file_path) and raise_error:
            raise NoFilesFoundError(
                msg=f"Cannot delete {file_path}. File does not exist",
                source=remove_files.__name__,
            )
        elif not os.path.isfile(file_path):
            pass
        else:
            try:
                os.remove(file_path)
            except:
                if raise_error:
                    raise PermissionError(
                        msg=f"Cannot read {file_path}. Is the file open in an alternative app?",
                        source=remove_files.__name__,
                    )
                else:
                    pass


def web_callback(url: str) -> None:
    try:
        result = urlparse(url)
        webbrowser.open_new(url)
        # return all([result.scheme, result.netloc])
    except ValueError:
        raise InvalidInputError(msg="Invalid URL: {url}", source=web_callback.__name__)


def get_pkg_version(pkg: str):
    """
    Helper to get the version of a package in the current python environment.

    :example:
    >>> get_pkg_version(pkg='simba-uw-tf-dev')
    >>> 1.82.7
    >>> get_pkg_version(pkg='bla-bla')
    >>> None
    """
    try:
        return pkg_resources.get_distribution(pkg).version
    except pkg_resources.DistributionNotFound:
        return None



__author__ = "Simon Nilsson"

import io
import os
from typing import Dict, List

import cv2
import matplotlib.pyplot as plt
import numpy as np
import PIL

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.data import detect_bouts
from simba.utils.enums import Formats
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.lookups import get_named_colors
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class GanttCreatorSingleProcess(ConfigReader, PlottingMixin):
    """
    Create gantt chart videos and/or images using a single core.

    .. note::
       `GitHub gantt tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#gantt-plot>`__.
       For improved run-time, see :meth:`simba.gantt_creator_mp.GanttCreatorMultiprocess` for multiprocess class.

    .. image:: _static/img/gantt_plot.png
       :width: 300
       :align: center


    :param str config_path: path to SimBA project config file in Configparser format.
    :param bool frame_setting: If True, creates individual frames.
    :param bool last_frm_setting: If True, creates single .png image representing entire video.
    :param bool video_setting: If True, creates videos
    :param dict style_attr: Attributes of gannt chart (size, font size, font rotation etc).
    :param List[str] files_found: File paths representing files with machine predictions e.g., ['project_folder/csv/machine_results/My_results.csv']

    Examples
    ----------
    >>> style_attr = {'width': 640, 'height': 480, 'font size': 12, 'font rotation': 45}
    >>> gantt_creator = GanttCreatorSingleProcess(config_path='tests/test_data/multi_animal_dlc_two_c57/project_folder/project_config.ini', frame_setting=False, video_setting=True, files_found=['tests/test_data/multi_animal_dlc_two_c57/project_folder/csv/machine_results/Together_1.csv'])
    >>> gantt_creator.run()

    """

    def __init__(
        self,
        config_path: str,
        frame_setting: bool,
        video_setting: bool,
        last_frm_setting: bool,
        files_found: List[str],
        style_attr: Dict[str, int],
    ):
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        (
            self.frame_setting,
            self.video_setting,
            self.style_attr,
            self.last_frm_setting,
        ) = (frame_setting, video_setting, style_attr, last_frm_setting)
        self.files_found = files_found
        if (
            (self.frame_setting != True)
            and (self.video_setting != True)
            and (self.last_frm_setting != True)
        ):
            raise NoSpecifiedOutputError(
                msg="SIMBA ERROR: Please select gantt videos, frames, and/or last frame."
            )
        check_if_filepath_list_is_empty(
            filepaths=self.files_found,
            error_msg="SIMBA ERROR: Zero files found in the project_folder/csv/machine_results directory. Create classification results before visualizing gantt charts",
        )
        self.colours = get_named_colors()
        self.colour_tuple_x = list(np.arange(3.5, 203.5, 5))
        if not os.path.exists(self.gantt_plot_dir):
            os.makedirs(self.gantt_plot_dir)
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        print("Processing {} video(s)...".format(str(len(self.files_found))))

    def run(self):
        for file_cnt, file_path in enumerate(self.files_found):
            _, self.video_name, _ = get_fn_ext(file_path)
            self.data_df = read_df(file_path, self.file_type).reset_index(drop=True)
            self.video_info_settings, _, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.bouts_df = detect_bouts(
                data_df=self.data_df, target_lst=self.clf_names, fps=self.fps
            )
            if self.frame_setting:
                self.save_frame_folder_dir = os.path.join(
                    self.gantt_plot_dir, self.video_name
                )
                if not os.path.exists(self.save_frame_folder_dir):
                    os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                self.save_video_path = os.path.join(
                    self.gantt_plot_dir, self.video_name + ".mp4"
                )
                self.writer = cv2.VideoWriter(
                    self.save_video_path,
                    self.fourcc,
                    self.fps,
                    (self.style_attr["width"], self.style_attr["height"]),
                )

            if self.last_frm_setting:
                _ = self.make_gantt_plot(
                    data_df=self.data_df,
                    bouts_df=self.bouts_df,
                    clf_names=self.clf_names,
                    fps=self.fps,
                    style_attr=self.style_attr,
                    video_name=self.video_name,
                    save_path=os.path.join(
                        self.gantt_plot_dir, self.video_name + "_final_image.png"
                    ),
                )

            if self.frame_setting or self.video_setting:
                for image_cnt, k in enumerate(range(len(self.data_df))):
                    fig, ax = plt.subplots()
                    relevant_rows = self.bouts_df.loc[self.bouts_df["End_frame"] <= k]
                    for i, event in enumerate(relevant_rows.groupby("Event")):
                        for x in self.clf_names:
                            if event[0] == x:
                                ix = self.clf_names.index(x)
                                data_event = event[1][["Start_time", "Bout_time"]]
                                ax.broken_barh(
                                    data_event.values,
                                    (self.colour_tuple_x[ix], 3),
                                    facecolors=self.colours[ix],
                                )

                    x_ticks_locs = x_lbls = np.round(
                        np.linspace(0, round((image_cnt / self.fps), 3), 6)
                    )
                    ax.set_xticks(x_ticks_locs)
                    ax.set_xticklabels(x_lbls)
                    ax.set_ylim(0, self.colour_tuple_x[len(self.clf_names)])
                    ax.set_yticks(np.arange(5, 5 * len(self.clf_names) + 1, 5))
                    ax.set_yticklabels(
                        self.clf_names, rotation=self.style_attr["font rotation"]
                    )
                    ax.tick_params(axis="both", labelsize=self.style_attr["font size"])
                    plt.xlabel("Session (s)", fontsize=self.style_attr["font size"])
                    ax.yaxis.grid(True)
                    buffer_ = io.BytesIO()
                    plt.savefig(buffer_, format="png")
                    buffer_.seek(0)
                    image = PIL.Image.open(buffer_)
                    ar = np.asarray(image)
                    open_cv_image = cv2.cvtColor(ar, cv2.COLOR_RGB2BGR)
                    open_cv_image = cv2.resize(
                        open_cv_image,
                        (self.style_attr["width"], self.style_attr["height"]),
                    )
                    frame = np.uint8(open_cv_image)
                    buffer_.close()
                    plt.close(fig)
                    if self.frame_setting:
                        frame_save_path = os.path.join(
                            self.save_frame_folder_dir, str(k) + ".png"
                        )
                        cv2.imwrite(frame_save_path, frame)
                    if self.video_setting:
                        self.writer.write(frame)
                    print(
                        "Gantt frame: {} / {}. Video: {} ({}/{})".format(
                            str(image_cnt + 1),
                            str(len(self.data_df)),
                            self.video_name,
                            str(file_cnt + 1),
                            len(self.files_found),
                        )
                    )
                if self.video_setting:
                    self.writer.release()
                print("Gantt for video {} saved...".format(self.video_name))
        self.timer.stop_timer()
        stdout_success(
            msg="All gantt visualizations created in project_folder/frames/output/gantt_plots directory",
            elapsed_time=self.timer.elapsed_time_str,
        )


# style_attr = {'width': 640, 'height': 480, 'font size': 12, 'font rotation': 45}
# test = GanttCreatorSingleProcess(config_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/project_config.ini',
#                                  frame_setting=False,
#                                  video_setting=True,
#                                  last_frm_setting=True,
#                                  style_attr=style_attr,
#                                  files_found=['/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals/project_folder/csv/outlier_corrected_movement_location/Testing_Video_3.csv'])
# test.create_gannt()


# style_attr = {'width': 640, 'height': 480, 'font size': 12, 'font rotation': 45}
# test = GanttCreatorSingleProcess(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                  frame_setting=False,
#                                  video_setting=True,
#                                  last_frm_setting=True,
#                                  style_attr=style_attr,
#                                  files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/targets_inserted/Together_1.csv'])
# test.run()



import argparse
import ast
import importlib.util
import inspect
import os
import subprocess
import sys
from pathlib import Path
from typing import Optional, Union

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_module_has_import, check_instance)
from simba.utils.errors import CountError, InvalidFileTypeError
from simba.utils.printing import stdout_warning
from simba.utils.read_write import get_fn_ext

ABSTRACT_CLASS_NAME = "AbstractFeatureExtraction"


class CustomFeatureExtractor(ConfigReader):
    """
    Class to execute a feature extraction process based on the user-defined Python script.

    This method performs the following steps:
    1. Parses the user-defined Python script.
    2. Identifies class and function names in the script.
    3. Checks for the presence of required imports and specific code patterns.
    4. Handle cases of multiple classes and missing configuration arguments.
    5. Invokes the feature extraction process if conditions are met.

    .. notes::

        `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/extractFeatures.md>`_.

        The user defined feature extraction class needs to contain a ``config_path`` init argument

        If the feature extraction class contains multiple classes, then the first class  is used.

        If the feature extraction class relies on argparse AND inherits ``simba.mixins.abstract_classes.AbstractFeatureExtraction``
        then the feature extraction class will be executed thtough the subprocess module. This ensures that the GUI cannot interfere
        with the feature extraction process, and reliably exexution of multicore processes if present in the feature-extraction class.

        If the feature extraction class does not rely on argparse, then the class will be loaded and executed in python through ``sys``. It
        might still be quick and reliable (it often is). However, I have noted that some function, in particular functions that rely on
        ``multiprocessing.imap`` are disrupted by the GUI with unacceptable effects on runtime when executed thriugh ``sys``.


    :example:
    >>> test = CustomFeatureExtractor(extractor_file_path='/simba/misc/piotr.py', config_path='/Users/simon/Desktop/envs/troubleshooting/piotr/project_folder/train-20231108-sh9-frames-with-p-lt-2_plus3-&3_best-f1.ini')
    >>> test.run()
    >>> test = CustomFeatureExtractor(config_path='/Users/simon/Desktop/envs/troubleshooting/piotr/project_folder/train-20231108-sh9-frames-with-p-lt-2_plus3-&3_best-f1.ini', file_path='/simba/misc/piotr.py')
    >>> test.run()
    """

    def __init__(
        self,
        extractor_file_path: Union[str, os.PathLike],
        config_path: Union[str, os.PathLike],
    ):
        check_file_exist_and_readable(file_path=config_path)
        check_file_exist_and_readable(file_path=extractor_file_path)
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)

        file_dir, file_name, file_extension = get_fn_ext(filepath=extractor_file_path)
        if file_extension != ".py":
            raise InvalidFileTypeError(
                msg=f"The user-defined feature extraction file ({extractor_file_path}) is not a .py file-extension",
                source=self.__class__.__name__,
            )
        self.extractor_file_path, self.config_path = extractor_file_path, config_path

    def _find_astClass_from_name(
        self, parsed_py: ast.Module, class_name: str
    ) -> ast.ClassDef:
        """
        Find and return the AST node representing a class with the specified name in the given AST module.

        :param ast.Module parsed_py: The AST module to search for the class.
        :param str class_name: The name of the class to find.
        :return ast.ClassDef : The AST node representing the class with the specified name.
        """

        classes = [n for n in parsed_py.body if isinstance(n, ast.ClassDef)]
        classes_match = [n for n in classes if n.name is class_name]
        if len(classes_match) == 0:
            raise CountError(
                msg=f"No classes named {class_name} in {self.extractor_file_path}",
                source=self.__class__.__name__,
            )
        else:
            return classes_match[0]

    def _find_class_names(self, parsed_py: ast.Module) -> list:
        """
        Find and return the names of classes defined in the given AST module.

        :param ast.Module parsed_py: The AST module to search for function names.
        :return list: A list of class names found in the AST module.
        """
        classes = [n for n in parsed_py.body if isinstance(n, ast.ClassDef)]
        return [x.name for x in classes]

    def _find_function_names(self, parsed_py: ast.Module) -> list:
        """
        Find and return the names of functions defined in the given AST module.

        :param ast.Module parsed_py: The AST module to search for function names.
        :return list: A list of function names found in the AST module.
        """
        check_instance(
            source=self.__class__.__name__,
            instance=parsed_py,
            accepted_types=ast.Module,
        )
        functions = [n for n in parsed_py.body if isinstance(n, ast.FunctionDef)]
        return [x.name for x in functions]

    def _check_inheritance(
        self, class_: ast.ClassDef, inheritance: Optional[str] = ABSTRACT_CLASS_NAME
    ) -> bool:
        """
        Check if a class inherits from a specified class.

        :param ast.ClassDef class_: The AST node representing the class to check.
        :param Optional[str] inheritance: The name of the class to check for inheritance. Defaults to the abstract class name.
        :return bool: True if the specified class is found in the class's bases, False otherwise.
        """
        check_instance(
            source=self.__class__.__name__, instance=class_, accepted_types=ast.ClassDef
        )
        bases = [base.id for base in class_.bases if isinstance(base, ast.Name)]
        if inheritance in bases:
            return True
        else:
            return False

    def has_block(self, file_path: Union[str, os.PathLike], target: str) -> bool:
        """
        Check if a specified block of text exists in a file.

        :param Union[str, os.PathLike] file_path: Path to the file to check.
        :param str target: The block of text to search for in the file.
        :return bool: True if the specified block of text is found in the file, False otherwise.
        """

        with open(file_path, "r") as file:
            lines = file.readlines()
            for line in lines:
                if line.strip() == target:
                    return True
        return False

    def has_argparser_argument(
        self, file_path: Union[str, os.PathLike], target_argument: str
    ) -> bool:
        """
        Check if a Python script, using argparse, has a specific command-line argument.

        :param Union[str, os.PathLike] file_path: Path to the Python script file.
        :param str target_argument: The name of the argument to check.
        :return bool: True if the specified argument is found in the script's argparse configuration, False otherwise.
        """

        parser = argparse.ArgumentParser()
        parser.add_argument("--" + target_argument)
        file_content = Path(file_path).read_text()
        try:
            parsed_args, _ = parser.parse_known_args(file_content.split())
            return hasattr(parsed_args, target_argument)
        except argparse.ArgumentError:
            return False

    def run(self):
        parsed_py = ast.parse(Path(self.extractor_file_path).read_text())
        class_names = self._find_class_names(parsed_py=parsed_py)
        function_names = self._find_function_names(parsed_py=parsed_py)
        has_argparse = check_if_module_has_import(
            parsed_file=parsed_py, import_name="argparse"
        )
        if len(class_names) < 1:
            raise CountError(
                msg=f"The user-defined feature extraction file ({self.extractor_file_path}) contains no python classes",
                source=self.__class__.__name__,
            )
        elif len(class_names) > 1:
            stdout_warning(
                msg=f"The user-defined feature extraction file ({self.extractor_file_path}) contains more than 1 python class. SimBA will use the first python class: {class_names[0]}."
            )
        class_name = class_names[0]
        class_ = self._find_astClass_from_name(
            parsed_py=parsed_py, class_name=class_name
        )
        has_abstract_inheritance = self._check_inheritance(class_=class_)
        has_main_block = self.has_block(
            file_path=self.extractor_file_path, target='if __name__ == "__main__":'
        )
        has_config_argparse = self.has_argparser_argument(
            file_path=self.extractor_file_path, target_argument="config_path"
        )
        if (
            has_abstract_inheritance
            and has_config_argparse
            and has_argparse
            and has_main_block
        ):
            command = f'python "{self.extractor_file_path}" --config_path "{self.config_path}"'
            print(
                "Follow feature extraction progress in the operating system terminal window..."
            )
            subprocess.call(command, shell=True)

        elif not has_config_argparse and not has_argparse:
            spec = importlib.util.spec_from_file_location(
                class_name, self.extractor_file_path
            )
            user_module = importlib.util.module_from_spec(spec)
            sys.modules[class_name] = user_module
            spec.loader.exec_module(user_module)
            user_class = getattr(user_module, class_name)

            if has_abstract_inheritance:
                feature_extractor = user_class(self.config_path)
                feature_extractor.run()

            else:
                user_class(self.config_path)


# test = UserDefinedFeatureExtractor(extractor_file_path='/Users/simon/Desktop/envs/simba_dev/simba/feature_extractors/misc/piotr.py',
#                             config_path='/Users/simon/Desktop/envs/troubleshooting/piotr/project_folder/train-20231108-sh9-frames-with-p-lt-2_plus3-&3_best-f1.ini')
# test.run()



__author__ = "Simon Nilsson"


import glob
import os
from copy import deepcopy

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_if_dir_exists,
                                check_if_filepath_list_is_empty)
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df


class DeepEthogramImporter(ConfigReader):
    """
    Append DeepEthogram optical flow annotations onto featurized pose-estimation data.

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_dir: path to folder holding DeepEthogram data files is CSV format

    .. note::
       `Third-party import tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/third_party_annot.md>`__.
       `Example expected input <https://github.com/sgoldenlab/simba/blob/master/misc/deep_ethogram_labels.csv>`__.

    Examples
    ----------
    >>> deepethogram_importer = DeepEthogramImporter(config_path=r'MySimBAConfigPath', data_dir=r'MyDeepEthogramDir')
    >>> deepethogram_importer.run()

    References
    ----------

    .. [1] `DeepEthogram repo <https://github.com/jbohnslav/deepethogram>`__.
    .. [2] `Example DeepEthogram input file <https://github.com/sgoldenlab/simba/blob/master/misc/deep_ethogram_labels.csv>`__.
    """

    def __init__(self, data_dir: str, config_path: str):
        super().__init__(config_path=config_path)
        self.data_dir = data_dir
        check_if_dir_exists(in_dir=self.data_dir)
        self.deepethogram_files_found = glob.glob(self.data_dir + "/*.csv")
        check_if_filepath_list_is_empty(
            filepaths=self.deepethogram_files_found,
            error_msg=f"SIMBA ERROR: ZERO DeepEthogram CSV files found in {self.data_dir} directory",
        )
        check_if_filepath_list_is_empty(
            filepaths=self.feature_file_paths,
            error_msg="SIMBA ERROR: ZERO files found in the project_folder/csv/features_extracted directory",
        )
        feature_file_names, self.matches_dict = [], {}
        for feature_file_path in self.feature_file_paths:
            _, file_name, ext = get_fn_ext(feature_file_path)
            feature_file_names.append(file_name)
        for file_path in self.deepethogram_files_found:
            _, file_name, ext = get_fn_ext(file_path)
            if file_name in feature_file_names:
                self.matches_dict[file_path] = os.path.join(
                    self.features_dir, file_name + ext
                )
                pass
            elif file_name.endswith("_labels"):
                short_file_name = file_name[:-7]
                if short_file_name in feature_file_names:
                    self.matches_dict[file_path] = os.path.join(
                        self.features_dir, short_file_name + ext
                    )
                    pass
            else:
                print(
                    "SIMBA ERROR: Could not find file in project_folder/csv/features_extracted directory representing {}".format(
                        file_name
                    )
                )
                raise FileNotFoundError()

    def run(self):
        for cnt, (k, v) in enumerate(self.matches_dict.items()):
            _, video_name, _ = get_fn_ext(filepath=v)
            self.annotations_df = read_df(
                file_path=k, file_type=self.file_type
            ).reset_index(drop=True)
            self.features_df = read_df(
                file_path=v, file_type=self.file_type
            ).reset_index(drop=True)
            _, _, self.fps = self.read_video_info(video_name=video_name)
            for clf_name in self.clf_names:
                if clf_name not in self.annotations_df.columns:
                    print(
                        "SIMBA ERROR: No annotations for behavior {} found in DeepEthogram annotation file for video {}"
                        "Exclude {} from your SimBA project or add DeepEthogram annotations for {} for video {}.".format(
                            clf_name, video_name, clf_name, clf_name, video_name
                        )
                    )
                    raise ValueError()
            if len(self.annotations_df) > len(self.features_df):
                print(
                    f"SIMBA WARNING: The DEEPETHOGRAM annotations for video {video_name} contain data for {str(len(self.annotations_df))} frames. The pose-estimation features for the same video contain data for {str(len(self.features_df))} frames. "
                    "SimBA will use the annotations for the frames present in the pose-estimation data and discard the rest."
                )
                self.annotations_df = self.annotations_df.head(len(self.features_df))
            if len(self.annotations_df) < len(self.features_df):
                print(
                    f"SIMBA WARNING: The DEEPETHOGRAM annotations for video {video_name} contain data for {str(len(self.annotations_df))} frames. The pose-estimation features for the same video contain data for {str(len(self.features_df))} frames. "
                    "SimBA expects the annotations and pose-estimation data to contain an equal number of frames. SimBA will assume that "
                    "the un-annotated frames have no behaviors present."
                )
                padding = pd.DataFrame(
                    [[0] * (len(self.features_df) - len(self.annotations_df))],
                    columns=self.annotations_df,
                )
                self.annotations_df = self.annotations_df.append(
                    padding, ignore_index=True
                )

            self.out_data = deepcopy(self.features_df)
            for clf_name in self.clf_names:
                self.out_data[clf_name] = self.annotations_df[clf_name]

            save_path = os.path.join(
                self.targets_folder, video_name + "." + self.file_type
            )
            write_df(df=self.out_data, file_type=self.file_type, save_path=save_path)
            print("DeepEthogram annotation for video {} saved...".format(video_name))

        stdout_success(
            msg=f"Annotations for {str(len(list(self.clf_names)))} behaviors added to {len(self.matches_dict.keys())} videos and saved in the project_folder/csv/targets_inserted directory."
        )


# test = DeepEthogramImporter(deep_ethogram_dir='/Users/simon/Desktop/troubleshooting/deepethnogram/deepethnogram',
#                             config_path='/Users/simon/Desktop/troubleshooting/deepethnogram/project_folder/project_config.ini')
# test.run()



__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import List, Optional, Union

import pandas as pd

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_file_exist_and_readable, check_if_filepath_list_is_empty)
from simba.utils.data import detect_bouts
from simba.utils.enums import Options, TagNames
from simba.utils.errors import InvalidInputError
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df
from simba.utils.warnings import NoDataFoundWarning


class AggregateClfCalculator(ConfigReader):
    """
    Compute aggregate descriptive statistics from classification data.

    :param str config_path: path to SimBA project config file in Configparser format
    :param List[str] data_measures: Aggregate statistics measures to calculate. OPTIONS: ['Bout count', 'Total event duration (s)',
        'Mean event bout duration (s)', 'Median event bout duration (s)', 'First event occurrence (s)',
        'Mean event bout interval duration (s)', 'Median event bout interval duration (s)']
    :param List[str] classifiers: Classifiers to calculate aggregate statistics for. E.g.,: ['Attack', 'Sniffing']
    :param Optional[List[str]] video_meta_data: Video metadata to include in the output. Options: 'Frame count', 'Video length (s)'.
    :param bool detailed_bout_data: If True, save detailed data for each bout in each video (start frame, end frame, bout time etc.)
    :param bool transpose: If True, then one video per row. Else, one meassure per row. Default: False.

    .. note::
       `GitHub tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md#part-4--analyze-machine-results>`__.

       `Example expected ouput file <https://github.com/sgoldenlab/simba/blob/master/misc/detailed_bout_data_summary_20231011091832.csv>`__.

    Examples
    -----
    >>> clf_log_creator = AggregateClfCalculator(config_path="MyConfigPath", data_measures=['Bout count', 'Total event duration (s)'], classifiers=['Attack', 'Sniffing'])
    >>> clf_log_creator.run()
    >>> clf_log_creator.save()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data_measures: List[
            Literal[
                "Bout count",
                "Total event duration (s)",
                "Mean event bout duration (s)",
                "Median event bout duration (s)",
                "First event occurrence (s)",
                "Mean event bout interval duration (s)",
                "Median event bout interval duration (s)",
            ]
        ],
        classifiers: List[str],
        detailed_bout_data: bool = False,
        transpose: bool = False,
        video_meta_data: Optional[
            List[Literal["Frame count", "Video length (s)"]]
        ] = None,
    ):

        super().__init__(config_path=config_path)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        (
            self.chosen_measures,
            self.classifiers,
            self.video_meta_data,
            self.detailed_bout_data,
        ) = (data_measures, classifiers, video_meta_data, detailed_bout_data)
        if (
            len(
                list(
                    set(self.chosen_measures).intersection(
                        Options.CLF_DESCRIPTIVES_OPTIONS.value
                    )
                )
            )
            == 0
        ):
            raise InvalidInputError(
                msg=f"No valid input data measures. OPTIONS: {Options.CLF_DESCRIPTIVES_OPTIONS.value}",
                source=self.__class__.__name__,
            )

        self.save_path = os.path.join(
            self.project_path, "logs", f"data_summary_{self.datetime}.csv"
        )
        self.detailed_save_path = os.path.join(
            self.project_path, "logs", f"detailed_bout_data_summary_{self.datetime}.csv"
        )
        self.transpose = transpose
        check_if_filepath_list_is_empty(
            filepaths=self.machine_results_paths,
            error_msg="SIMBA ERROR: No data files found in the project_folder/csv/machine_results directory. Run classifiers before analysing results.",
        )
        print(
            f"Analyzing {str(len(self.machine_results_paths))} file(s) for {str(len(self.clf_names))} classifiers..."
        )

    def run(self):
        self.results_df, self.bouts_df_lst = pd.DataFrame(), []
        check_all_file_names_are_represented_in_video_log(
            video_info_df=self.video_info_df, data_paths=self.machine_results_paths
        )
        for file_cnt, file_path in enumerate(self.machine_results_paths):
            _, file_name, _ = get_fn_ext(file_path)
            print("Analyzing video {}...".format(file_name))
            _, _, fps = self.read_video_info(video_name=file_name)
            check_file_exist_and_readable(file_path)
            data_df = read_df(file_path, self.file_type)
            bouts_df = detect_bouts(data_df=data_df, target_lst=self.clf_names, fps=fps)
            if self.detailed_bout_data and (len(bouts_df) > 0):
                bouts_df_for_detailes = deepcopy(bouts_df)
                bouts_df_for_detailes.insert(loc=0, column="Video", value=file_name)
                self.bouts_df_lst.append(bouts_df_for_detailes)
            bouts_df["Shifted start"] = bouts_df["Start_time"].shift(-1)
            bouts_df["Interval duration"] = (
                bouts_df["Shifted start"] - bouts_df["End Time"]
            )
            for clf in self.clf_names:
                clf_results_dict = {}
                clf_data = bouts_df.loc[bouts_df["Event"] == clf]
                if len(clf_data) > 0:
                    clf_results_dict["First event occurrence (s)"] = round(
                        clf_data["Start_time"].min(), 3
                    )
                    clf_results_dict["Bout count"] = len(clf_data)
                    clf_results_dict["Total event duration (s)"] = round(
                        clf_data["Bout_time"].sum(), 3
                    )
                    clf_results_dict["Mean event bout duration (s)"] = round(
                        clf_data["Bout_time"].mean(), 3
                    )
                    clf_results_dict["Median event bout duration (s)"] = round(
                        clf_data["Bout_time"].median(), 3
                    )
                else:
                    clf_results_dict["First event occurrence (s)"] = None
                    clf_results_dict["Bout count"] = None
                    clf_results_dict["Total event duration (s)"] = None
                    clf_results_dict["Mean event bout duration (s)"] = None
                    clf_results_dict["Median event bout duration (s)"] = None
                if len(clf_data) > 1:
                    interval_df = clf_data[:-1].copy()
                    clf_results_dict["Mean event bout interval duration (s)"] = round(
                        interval_df["Interval duration"].mean(), 3
                    )
                    clf_results_dict["Median event bout interval duration (s)"] = round(
                        interval_df["Interval duration"].median(), 3
                    )
                else:
                    clf_results_dict["Mean event bout interval duration (s)"] = None
                    clf_results_dict["Median event bout interval duration (s)"] = None
                if self.video_meta_data is not None:
                    if "Frame count" in self.video_meta_data:
                        clf_results_dict["Frame count"] = len(data_df)
                    if "Video length (s)" in self.video_meta_data:
                        clf_results_dict["Video length (s)"] = round(
                            len(data_df) / fps, 3
                        )
                video_clf_pd = (
                    pd.DataFrame.from_dict(clf_results_dict, orient="index")
                    .reset_index()
                    .rename(columns={"index": "Measure", 0: "Value"})
                )
                video_clf_pd.insert(loc=0, column="Classifier", value=clf)
                video_clf_pd.insert(loc=0, column="Video", value=file_name)
                self.results_df = pd.concat([self.results_df, video_clf_pd], axis=0)

    def save(self) -> None:
        """
        Method to save classifier aggregate statistics created in :meth:`~simba.ClfLogCreator.analyze_data` to disk.
        Results are stored in the `project_folder/logs` directory of the SimBA project
        """

        self.results_df = (
            self.results_df[
                self.results_df["Measure"].isin(
                    self.chosen_measures + self.video_meta_data
                )
            ]
            .sort_values(by=["Video", "Classifier", "Measure"])
            .reset_index(drop=True)
        )
        self.results_df = self.results_df[
            self.results_df["Classifier"].isin(self.classifiers)
        ].set_index("Video")
        if not self.transpose:
            self.results_df.to_csv(self.save_path)
        else:
            self.results_df.loc[
                self.results_df["Measure"].isin(self.video_meta_data), "Classifier"
            ] = "Metadata"
            self.results_df = self.results_df.reset_index().drop_duplicates(
                subset=["Video", "Classifier", "Measure"], keep="first"
            )
            self.results_df["clf_measure"] = (
                self.results_df["Classifier"] + " - " + self.results_df["Measure"]
            )
            self.results_df = self.results_df.drop(
                ["Classifier", "Measure"], axis=1
            ).reset_index()
            self.results_df = self.results_df.pivot(
                index="Video", columns="clf_measure", values="Value"
            )
            self.results_df.to_csv(self.save_path)

        if self.detailed_bout_data:
            self.bouts_df = pd.concat(self.bouts_df_lst, axis=0)
            self.bouts_df = self.bouts_df[
                self.bouts_df["Event"].isin(self.classifiers)
            ].set_index("Video")
            if len(self.bouts_df) == 0:
                NoDataFoundWarning(
                    msg=f"No detailed bout data saved: No bouts detected for the selected classifiers: {self.classifiers}"
                )
            else:
                self.bouts_df.to_csv(self.detailed_save_path)
                stdout_success(
                    msg=f"Detailed bout data log saved at {self.detailed_save_path}",
                    source=self.__class__.__name__,
                )
        self.timer.stop_timer()
        stdout_success(
            msg=f"Data aggregate log saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# test = AggregateClfCalculator(config_path=r"/Users/simon/Desktop/envs/troubleshooting/raph/project_folder/project_config.ini",
#                               data_measures=['Total event duration (s)', 'Median event bout duration (s)'],
#                               classifiers=['walking'],
#                               video_meta_data =['Frame count'],
#                               transpose=True)
#
#
#
# test.run()
# test.save()



import numpy as np
import pandas as pd
from pathlib import Path
import pytest
import pickle
import os

from simba.mixins.statistics_mixin import Statistics
from simba.utils.enums import TestPaths, Paths


@pytest.fixture(scope="session")
def sample_data_dir():
    return Path(__file__).resolve().parent / "data" / "sample_data"

def test_rolling_independent_sample_t():
    data = np.array([1, 3, 4, 7, 10, 6, 10, 12, 239, 34, 21]).astype(np.float32)
    results = Statistics().rolling_independent_sample_t(data=data, time_window=1.0, fps=2)
    expected_results = np.array([-1. , -1. , -2.74562589, -2.74562589, -1.41421356, -1.41421356, -1.8973666 , -1.8973666 , -1.73146689, -1.73146689, 0.92005224])
    assert np.allclose(results, expected_results)


@pytest.mark.parametrize("critical_values, expected_results", [(True, (-2.5266047334683974, True)), (False, (-2.5266047334683974, None))])
def test_independent_samples_t(critical_values, expected_results):
    sample_1 = np.array([1, 2, 3, 1, 3, 2, 1, 10, 8, 4, 10]).astype(np.float32)
    sample_2 = np.array([2, 5, 10, 4, 8, 10, 7, 10, 7, 10, 10]).astype(np.float32)
    if critical_values:
        critical_values = pickle.load(open(Paths.CRITICAL_VALUES.value, 'rb'))['independent_t_test']['one_tail'].values
        results = Statistics().independent_samples_t(sample_1=sample_1, sample_2=sample_2, critical_values=critical_values)
    else:
        results = Statistics().independent_samples_t(sample_1=sample_1, sample_2=sample_2)
    assert results == expected_results

def test_cohens_d():
    sample_1 = np.array([2, 4, 7, 3, 7, 35, 8, 9]).astype(np.float64)
    sample_2 = np.array([4, 8, 14, 6, 14, 70, 16, 18]).astype(np.float64)
    results = Statistics().cohens_d(sample_1=sample_1, sample_2=sample_2)
    assert results == -0.5952099775170546

def test_rolling_cohens_d():
    data = np.array([11.66098032, 11.52882001,  8.28137675,  9.63358521, 10.55202715, 8.57905361, 14.06717964,  8.85201707]).astype(np.float64)
    results = Statistics().rolling_cohens_d(data=data, time_windows=np.array([1.0]), fps=4.0)
    expected_results = np.array([[0], [0], [0], [0], [-0.12864096], [-0.12864096], [-0.12864096], [-0.12864096]]).astype(np.float64)
    assert np.allclose(results, expected_results)

def test_rolling_two_sample_ks():
    data = np.array([4., 0., 4., 4., 2., 2., 3., 0., 0., 3., 0., 1., 0., 1., 2.], dtype=np.float32)
    expected_results = np.array([-1.,  1.,  1.,  0.,  1.,  0.,  1.,  1.,  0.,  1.,  1.,  1.,  1., 1.,  1.])
    results = Statistics().rolling_two_sample_ks(data=data, time_window=1, fps=1)
    assert np.allclose(results, expected_results)

def test_two_sample_ks():
    sample_1 = np.array([1, 2, 3, 1, 3, 2, 1, 10, 8, 4, 10]).astype(np.float32)
    sample_2 = np.array([10, 5, 10, 4, 8, 10, 7, 10, 7, 10, 10]).astype(np.float32)
    results = Statistics().two_sample_ks(sample_1=sample_1, sample_2=sample_2)
    assert results == (0.7272727272727273, None)

def test_one_way_anova():
    sample_1 = np.array([1, 2, 3, 1, 3, 2, 1, 10, 8, 4, 10])
    sample_2 = np.array([8, 5, 5, 8, 8, 9, 10, 1, 7, 10, 10])
    critical_values = pickle.load(open(Paths.CRITICAL_VALUES.value, 'rb'))['f']['one_tail'].values
    results = Statistics().one_way_anova(sample_1=sample_2, sample_2=sample_1, critical_values=critical_values)
    assert results == (5.8483754512635375, True)

def test_rolling_one_way_anova():
    sample = np.array([11.50357579,  9.32686126,  9.63068002,  9.6988871 , 11.40003894, 9.42987609,  9.28868644,  9.77685886, 11.19658191,  8.86118912], dtype=np.float32)
    results = Statistics().rolling_one_way_anova(data=sample, time_windows=np.array([1.0]), fps=2)
    expected_results = np.array([[0.        ], [0.        ], [0.47496008], [0.47496008], [0.57924093], [0.57924093], [0.75560836], [0.75560836], [0.17295212], [0.17295212]])
    assert np.allclose(results, expected_results)

@pytest.mark.parametrize('bucket_method, expected_results', [("fd", 0.59), ("doane", 0.37), ("auto", 0.53), ("scott", 1.22), ("stone", 0.03), ("rice", 0.53), ("sturges", 0.53), ("sqrt", 0.66), ("auto",   0.53)])
def test_kullback_leibler_divergence(bucket_method, expected_results):
    sample_1 = np.array([1, 2, 3, 1, 3, 2, 1, 10, 8, 4, 10])
    sample_2 = np.array([8, 5, 5, 8, 8, 9, 10, 1, 7, 10, 10])
    results = Statistics().kullback_leibler_divergence(sample_1=sample_1, sample_2=sample_2, fill_value=1, bucket_method=bucket_method)
    assert np.round(results, 2) == expected_results

@pytest.mark.parametrize('bucket_method, expected_results', [("fd", 0.13), ("doane", 0.32), ("auto", 0.3), ("scott", 0.25), ("stone", 0.43), ("rice", 0.3), ("sturges", 0.3), ("sqrt", 0.28)])
def test_jensen_shannon_divergence(bucket_method, expected_results):
    sample_1 = np.array([1, 2, 3, 1, 3, 2, 1, 10, 8, 4, 10])
    sample_2 = np.array([8, 5, 5, 8, 8, 9, 10, 1, 7, 10, 10])
    results = Statistics().jensen_shannon_divergence(sample_1=sample_1, sample_2=sample_2, bucket_method=bucket_method)
    assert np.round(results, 2) == expected_results


@pytest.mark.parametrize('bucket_method, expected_results', [("fd", 0.05), ("doane", 0.03), ("auto", 0.03), ("scott", 0.12), ("stone", 0.0), ("rice", 0.03), ("sturges", 0.03), ("sqrt", 0.04)])
def test_wasserstein_distance(bucket_method, expected_results):
    sample_1 = np.array([1, 2, 3, 1, 3, 2, 1, 10, 8, 4, 10])
    sample_2 = np.array([8, 5, 5, 8, 8, 9, 10, 1, 7, 10, 10])
    results = Statistics().wasserstein_distance(sample_1=sample_1, sample_2=sample_2, bucket_method=bucket_method)
    assert np.round(results, 2) == expected_results

@pytest.mark.parametrize('bucket_method, expected_results', [("fd", 1.13), ("doane", 0.72), ("auto", 1.0), ("scott", 2.26), ("stone", 0.06), ("rice", 1.0), ("sturges", 1.0), ("sqrt", 1.24)])
def test_population_stability_index(bucket_method, expected_results):
    sample_1 = np.array([1, 2, 3, 1, 3, 2, 1, 10, 8, 4, 10])
    sample_2 = np.array([8, 5, 5, 8, 8, 9, 10, 1, 7, 10, 10])
    results = Statistics().population_stability_index(sample_1=sample_1, sample_2=sample_2, bucket_method=bucket_method)
    assert np.round(results, 2) == expected_results

def test_kruskal_wallis():
    sample_1 = np.array([1, 1, 3, 4, 5]).astype(np.float64)
    sample_2 = np.array([6, 7, 8, 9, 10]).astype(np.float64)
    results = Statistics().kruskal_wallis(sample_1=sample_1, sample_2=sample_2)
    assert np.equal(results, 39.4)

def test_mann_whitney():
    sample_1 = np.array([1, 1, 3, 4, 5]).astype(np.float64)
    sample_2 = np.array([1, 3, 2, 1, 6]).astype(np.float64)
    results = Statistics().mann_whitney(sample_1=sample_1, sample_2=sample_2)
    assert np.equal(results, 11.5)

def test_brunner_munzel():
    sample_1 = np.array([1, 1, 3, 4, 5]).astype(np.float64)
    sample_2 = np.array([1, 3, 2, 1, 6]).astype(np.float64)
    results = Statistics().brunner_munzel(sample_1=sample_1, sample_2=sample_2)
    assert np.equal(np.round(results, 2), -0.31)

def test_pearsons_r():
    sample_1 = np.array([7, 2, 9, 4, 5, 6, 7, 8, 9]).astype(np.float32)
    sample_2 = np.array([1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5]).astype(np.float32)
    results = Statistics().pearsons_r(sample_1=sample_1, sample_2=sample_2)
    assert np.equal(np.round(results, 2), 0.47)

def test_spearman_rank_correlation():
    sample_1 = np.array([7, 2, 9, 4, 5, 6, 7, 8, 9]).astype(np.float32)
    sample_2 = np.array([1.5, 2.5, 3.5, 4.5, 5.5, 6.5, 7.5, 8.5, 9.5]).astype(np.float32)
    results = Statistics().spearman_rank_correlation(sample_1=sample_1, sample_2=sample_2)
    assert np.equal(np.round(results, 2), 0.53)

def test_rolling_mann_whitney():
    sample_1 = np.array([2., 0., 0., 1., 0., 1., 0., 0., 1., 1.], dtype=np.float32)
    results = Statistics().rolling_mann_whitney(data=sample_1, time_windows=np.array([1.0]), fps=1)
    expected_results = np.array([[0. ], [0. ], [0.5], [0. ], [0. ], [0. ], [0. ], [0.5], [0. ], [0.5]])
    assert np.allclose(results, expected_results)


def test_sliding_spearman_rank_correlation():
    sample_1 = np.array([9, 10, 13, 22, 15, 18, 15, 19, 32, 11]).astype(np.float32)
    sample_2 = np.array([11, 12, 15, 19, 21, 26, 19, 20, 22, 19]).astype(np.float32)
    expected_results = np.array([[ 0. ], [ 0. ], [ 0. ], [ 0. ], [ 0.9], [ 0.7], [ 0.3], [-0.5], [ 0.3], [ 0.6]])
    results = Statistics().sliding_spearman_rank_correlation(sample_1=sample_1, sample_2=sample_2, time_windows=np.array([0.5]), fps=10)
    assert np.allclose(results, expected_results)

def test_sliding_autocorrelation():
    data = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 1, 10, 11, 12, 13, 14]).astype(np.float32)
    expected_results = np.array([ 0.        ,  0.        ,  0.        ,  0.        ,  0.        , 0.        ,  0.        ,  0.        ,  0.        , -3.68623543,-2.02938271, -1.32312334, -1.75346696, -3.8074882 , -4.63441181])
    results = Statistics().sliding_autocorrelation(data=data, max_lag=0.5, time_window=1.0, fps=10)
    np.allclose(results, expected_results)

def test_kendall_tau():
    sample_1 = np.array([4, 2, 3, 4, 5, 7]).astype(np.float32)
    sample_2 = np.array([1, 2, 3, 4, 5, 7]).astype(np.float32)
    results = Statistics().kendall_tau(sample_1=sample_1, sample_2=sample_2)
    assert results == (0.7333333333333333, 2.0665401605809928)

def test_sliding_kendall_tau():
    sample_1 = np.array([4, 2, 3, 4, 5, 7]).astype(np.float32)
    sample_2 = np.array([1, 2, 3, 4, 5, 7]).astype(np.float32)
    time_windows = np.array([1.0, 1.5])
    expected_results = np.array([[ 0.        ,  0.        ],[ 0.        ,  0.        ],[-1.        ,  0.        ],[ 1.        , -0.33333333],[ 1.        ,  1.        ],[ 1.        ,  1.        ]])
    results = Statistics().sliding_kendall_tau(sample_1=sample_1, sample_2=sample_2, time_windows=time_windows, fps=2)
    assert np.allclose(expected_results, results)

@pytest.mark.parametrize('sample_data, expected_results', [('8x100_f32.npy', (8, ))])
def test_local_outlier_factor(sample_data, expected_results, sample_data_dir):
    data_path = os.path.join(sample_data_dir, sample_data)
    data = np.load(data_path)
    results = Statistics().local_outlier_factor(data=data, k=5)
    assert results.shape == expected_results
    assert np.issubdtype(results.dtype, np.number)

@pytest.mark.parametrize('sample_data, expected_results', [('8x100_f32.npy', (8, ))])
def test_hbos(sample_data, expected_results, sample_data_dir):
    data_path = os.path.join(sample_data_dir, sample_data)
    data = np.load(data_path)
    results = Statistics().hbos(data=data)
    assert results.shape == expected_results
    assert np.issubdtype(results.dtype, np.number)

def test_cohens_h():
    sample_1 = np.array([1, 0, 0, 1])
    sample_2 = np.array([1, 1, 1, 0])
    result = Statistics().cohens_h(sample_1=sample_1, sample_2=sample_2)
    assert np.round(result, 2) == -0.52

@pytest.mark.parametrize('time_windows', [(1.0, 2.0), (1.0), (1.5)])
def test_sliding_skew(time_windows):
    time_windows = np.array([time_windows]).flatten()
    data = np.random.randint(0, 100, (10,))
    results = Statistics().sliding_skew(data=data.astype(np.float32), time_windows=time_windows, sample_rate=2)
    assert np.issubdtype(results.dtype, np.number)
    assert results.shape[1] == time_windows.shape[0]

@pytest.mark.parametrize('time_windows', [(1.0, 2.0), (1.0), (1.5)])
def test_sliding_kurtosis(time_windows):
    time_windows = np.array([time_windows]).flatten()
    data = np.random.randint(0, 100, (10,))
    results = Statistics().sliding_kurtosis(data=data.astype(np.float32), time_windows=time_windows, sample_rate=2)
    assert np.issubdtype(results.dtype, np.number)
    assert results.shape[1] == time_windows.shape[0]

@pytest.mark.parametrize('k, data_size, medians', [(3, 100, True), (10, 200, False)])
def test_kmeans_1d(k, data_size, medians):
    data = np.random.randint(0, 100, (data_size,)).astype(np.float64)
    centroids, labels, median_res = Statistics().kmeans_1d(data=data, k=k, max_iters=10, calc_medians=medians)
    assert np.unique(centroids).shape[0] == k
    assert np.unique(labels).shape[0] == k
    assert labels.shape[0] == data_size
    if medians:
        assert len(list(median_res.keys())) == k

def test_hamming_distance():
    x, y = np.random.randint(0, 2, (10,)).astype(np.int8), np.random.randint(0, 2, (10,)).astype(np.int8)
    results = Statistics().hamming_distance(x=x, y=y)
    assert type(results) == float
    assert 0.0 <= results <= 1.0

@pytest.mark.parametrize('data_size,', [(100, 23), (50, 100), (1000, 10)])
def test_bray_curtis_dissimilarity(data_size):
    x = np.random.random(size=data_size).astype(np.float32)
    results = Statistics().bray_curtis_dissimilarity(x=x)
    assert np.issubdtype(results.dtype, np.number)
    assert results.shape[1] == data_size[0]
    assert results.shape[0] == data_size[0]

@pytest.mark.parametrize('bucket_method, data_size,', [("fd", 100), ("doane", 10), ("auto", 50), ("scott", 100), ("stone", 232), ("rice", 100), ("sturges", 20), ("sqrt", 40)])
def test_hellinger_distance(bucket_method, data_size):
    x = np.random.randint(-100, 100, (data_size))
    y = np.random.randint(-100, 100, (data_size))
    results = Statistics().hellinger_distance(x=x, y=y, bucket_method=bucket_method)
    assert type(results) == float


__author__ = "Simon Nilsson"

import os
from typing import List

import cv2
import numpy as np
import pandas as pd
from numba import jit, prange

from simba.mixins.config_reader import ConfigReader
from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.enums import Formats, TagNames
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class HeatmapperLocationSingleCore(ConfigReader, PlottingMixin):
    """
    Create heatmaps representing the location where animals spend time.
    For improved run-time, see :meth:`simba.heat_mapper_location_mp.HeatMapperLocationMultiprocess` for multiprocess class.

    ..note::
       `GitHub visualizations tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. image:: _static/img/heatmap_location.gif
       :width: 500
       :align: center

    :param str config_path: path to SimBA project config file in Configparser format
    :param str bodypart: The name of the body-part used to infer the location of the animal.
    :param int bin_size: The rectangular size of each heatmap location in millimeters. For example, `50` will divide the video frames into 5 centimeter rectangular spatial bins.
    :param str palette:  Heatmap pallette. Eg. 'jet', 'magma', 'inferno','plasma', 'viridis', 'gnuplot2'
    :param int or 'auto' max_scale: The max value in the heatmap in seconds. E.g., with a value of `10`, if the classified behavior has occured >= 10s within a rectangular bins, it will be filled with the same color.
    :param bool final_img_setting: If True, create a single image representing the last frame of the input video
    :param bool video_setting: If True, then create a video of heatmaps.
    :param bool frame_setting: If True, then create individual heatmap frames.

    :example:
    >>> style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'}
    >>> heatmapper = HeatmapperLocationSingleCore(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', style_attr = style_attr, final_img_setting=True, video_setting=True, frame_setting=False, bodypart='Nose_1', files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'])
    >>> heatmapper.run()
    """

    def __init__(
        self,
        config_path: str,
        bodypart: str,
        style_attr: dict,
        final_img_setting: bool,
        video_setting: bool,
        frame_setting: bool,
        files_found: List[str],
    ):

        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        log_event(
            logger_name=str(__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        if (not frame_setting) and (not video_setting) and (not final_img_setting):
            raise NoSpecifiedOutputError(
                msg="Please choose to select either heatmap videos, frames, and/or final image.",
                source=self.__class__.__name__,
            )

        self.frame_setting, self.video_setting = frame_setting, video_setting
        self.final_img_setting, self.bp = final_img_setting, bodypart
        self.style_attr = style_attr
        if not os.path.exists(self.heatmap_location_dir):
            os.makedirs(self.heatmap_location_dir)
        self.files_found = files_found
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.bp_lst = [self.bp + "_x", self.bp + "_y"]
        print(
            "Processing heatmaps for {} video(s)...".format(str(len(self.files_found)))
        )

    @staticmethod
    @jit(nopython=True)
    def __calculate_cum_array_final_img(loc_array: np.array):
        final_img = np.full((loc_array.shape[1], loc_array.shape[2]), 0)
        for frm in range(loc_array.shape[0]):
            for row in range(loc_array.shape[1]):
                for col in range(loc_array.shape[2]):
                    final_img[row, col] += loc_array[frm, row, col]
        return final_img

    def __calculate_bin_attr(
        self,
        data_df: pd.DataFrame,
        bp_lst: list,
        px_per_mm: int,
        img_width: int,
        img_height: int,
        bin_size: int,
        fps: int,
    ):

        bin_size_px = int(float(px_per_mm) * float(bin_size))
        horizontal_bin_cnt = int(img_width / bin_size_px)
        vertical_bin_cnt = int(img_height / bin_size_px)
        aspect_ratio = round((vertical_bin_cnt / horizontal_bin_cnt), 3)

        bp_data = data_df[bp_lst].to_numpy().astype(int)

        bin_dict = {}
        x_location, y_location = 0, 0
        for hbin in range(horizontal_bin_cnt):
            bin_dict[hbin] = {}
            for vbin in range(vertical_bin_cnt):
                bin_dict[hbin][vbin] = {
                    "top_left_x": x_location,
                    "top_left_y": y_location,
                    "bottom_right_x": x_location + bin_size_px,
                    "bottom_right_y": y_location + bin_size_px,
                }
                y_location += bin_size_px
            y_location = 0
            x_location += bin_size_px

        location_array = np.zeros(
            (bp_data.shape[0], vertical_bin_cnt, horizontal_bin_cnt)
        )

        for frm_cnt, frame in enumerate(bp_data):
            for h_bin_name, v_dict in bin_dict.items():
                for v_bin_name, c in v_dict.items():
                    if frame[0] <= c["bottom_right_x"] and frame[0] >= c["top_left_x"]:
                        if (
                            frame[1] <= c["bottom_right_y"]
                            and frame[0] >= c["top_left_y"]
                        ):
                            location_array[frm_cnt][v_bin_name][h_bin_name] = 1

        location_array = self.__calculate_cum_array(clf_array=location_array, fps=fps)

        return location_array, aspect_ratio

    @staticmethod
    @jit(nopython=True)
    def __calculate_cum_array(clf_array: np.array, fps: int):
        cum_sum_arr = np.full(clf_array.shape, np.nan)
        for frm_idx in prange(clf_array.shape[0]):
            frame_cum_sum = np.full((clf_array.shape[1], clf_array.shape[2]), 0.0)
            sliced_arr = clf_array[0:frm_idx]
            for i in range(sliced_arr.shape[0]):
                for j in range(sliced_arr.shape[1]):
                    for k in range(sliced_arr.shape[2]):
                        frame_cum_sum[j][k] += sliced_arr[i][j][k]
            cum_sum_arr[frm_idx] = frame_cum_sum

        return cum_sum_arr / fps

    def run(self):
        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, self.px_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.width, self.height = int(
                self.video_info["Resolution_width"].values[0]
            ), int(self.video_info["Resolution_height"].values[0])
            if self.video_setting:
                self.video_save_path = os.path.join(
                    self.heatmap_location_dir, self.video_name + ".mp4"
                )
                self.writer = cv2.VideoWriter(
                    self.video_save_path,
                    self.fourcc,
                    self.fps,
                    (self.width, self.height),
                )
            if self.frame_setting | self.final_img_setting:
                self.save_video_folder = os.path.join(
                    self.heatmap_location_dir, self.video_name
                )
                if not os.path.exists(self.save_video_folder):
                    os.makedirs(self.save_video_folder)
            self.data_df = read_df(file_path, self.file_type)[self.bp_lst]
            squares, aspect_ratio = GeometryMixin().bucket_img_into_grid_square(
                bucket_size_mm=self.style_attr["bin_size"],
                img_size=(self.width, self.height),
                px_per_mm=self.px_per_mm,
            )
            location_array = GeometryMixin().cumsum_geometries(
                data=self.data_df.values, fps=self.fps, geometries=squares
            )

            if self.style_attr["max_scale"] == "auto":
                self.max_scale = np.round(np.max(np.max(location_array[-1], axis=0)), 3)
                if self.max_scale == 0:
                    self.max_scale = 1
            else:
                self.max_scale = self.style_attr["max_scale"]

            if self.final_img_setting:
                self.make_location_heatmap_plot(
                    frm_data=location_array[-1:, :, :][0],
                    max_scale=self.max_scale,
                    palette=self.style_attr["palette"],
                    aspect_ratio=aspect_ratio,
                    file_name=os.path.join(
                        self.heatmap_location_dir, self.video_name + "_final_frm.png"
                    ),
                    shading=self.style_attr["shading"],
                    img_size=(self.width, self.height),
                    final_img=True,
                )
                print(
                    "Final heatmap image saved at {}".format(
                        os.path.join(self.save_video_folder, "_final_img.png")
                    )
                )

            if (self.frame_setting) or (self.video_setting):
                for frm_cnt, cumulative_frm in enumerate(
                    range(location_array.shape[0])
                ):
                    img = self.make_location_heatmap_plot(
                        frm_data=location_array[cumulative_frm, :, :],
                        max_scale=self.max_scale,
                        palette=self.style_attr["palette"],
                        aspect_ratio=aspect_ratio,
                        file_name=None,
                        shading=self.style_attr["shading"],
                        img_size=(self.width, self.height),
                        final_img=False,
                    )
                    if self.video_setting:
                        self.writer.write(img)
                    if self.frame_setting:
                        frame_save_path = os.path.join(
                            self.save_video_folder, str(frm_cnt) + ".png"
                        )
                        cv2.imwrite(frame_save_path, img)

                    print(
                        "Heatmap frame: {} / {}. Video: {} ({}/{})".format(
                            str(frm_cnt + 1),
                            str(len(self.data_df)),
                            self.video_name,
                            str(file_cnt + 1),
                            len(self.files_found),
                        )
                    )

            if self.video_setting:
                self.writer.release()
            video_timer.stop_timer()
            print(
                f"Heatmap plot for video {self.video_name} saved (elapsed time: {video_timer.elapsed_time_str}s"
            )
        self.timer.stop_timer()
        stdout_success(
            msg=f"Created heatmaps for {str(len(self.files_found))} videos",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


#
# test = HeatmapperLocationSingleCore(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 100, 'max_scale': 'auto'},
#                                       final_img_setting=True,
#                                       video_setting=True,
#                                       frame_setting=False,
#                                       bodypart='Nose_1',
#                                       files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'])
# test.run()


# test = HeatmapperLocationSingleCore(config_path='/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/project_config.ini',
#                                       style_attr = {'palette': 'jet', 'shading': 'gouraud', 'bin_size': 50, 'max_scale': 'auto'},
#                                       final_img_setting=False,
#                                       video_setting=True,
#                                       frame_setting=False,
#                                       bodypart='Nose',
#                                       files_found=['/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/csv/outlier_corrected_movement_location/PD1406_2022-05-24_RVDG_GCaMP8s-Gi_Video_Day_22_Baseline.csv'])
# test.create_heatmaps()


# test = HeatmapperLocation(config_path='/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini',
#                           bodypart='Nose_1',
#                           bin_size=50,
#                           palette='jet',
#                           max_scale='auto',
#                           final_img_setting=True,
#                           video_setting=False,
#                           frame_setting=False)
# test.create_heatmaps()
